"""
Dict helper module
"""


class NestedDict(dict):
    """
    A dictionary that automatically creates nested dictionaries for missing keys.

    This class extends the built-in `dict` to simplify working with deeply nested structures.
    If a key is accessed but does not exist, it will be created automatically with a new `NestedDict` as its value.

    Source: https://stackoverflow.com/a/5369984/10237506

    Example:
        >>> nd = NestedDict()
        >>> nd['a']['b']['c'] = 42
        >>> nd
        {'a': {'b': {'c': 42}}}

        >>> nd['x']['y']
        {}
    """

    __slots__ = ()

    def __getitem__(self, key):
        """
        Retrieve the value for a key, or create a nested dictionary for missing keys.

        Args:
            key (Hashable): The key to retrieve or create.

        Returns:
            Any: The value associated with the key, or a new `NestedDict` for missing keys.

        Example:
            >>> nd = NestedDict()
            >>> nd['foo']  # Creates a new NestedDict for 'foo'
            {}

        Note:
            If the key exists, its value is returned. Otherwise, a new `NestedDict` is created,
            stored, and returned.
        """
        if key in self: return self.get(key)
        return self.setdefault(key, NestedDict())


class DictWithLowerStore(dict):
    """
    A ``dict``-like object with a lower-cased key store.

    All keys are expected to be strings. The structure remembers the
    case of the lower-cased key to be set, and methods like ``get()``
    and ``get_key()`` will use the lower-cased store. However, querying
    and contains testing is case sensitive::

        dls = DictWithLowerStore()
        dls['Accept'] = 'application/json'
        dls['aCCEPT'] == 'application/json'         # False (raises KeyError)
        dls['Accept'] == 'application/json'         # True
        dls.get('aCCEPT') == 'application/json'     # True

        dls.get_key('aCCEPT') == 'Accept'           # True
        list(dls) == ['Accept']                     # True

    .. NOTE::
       I don't want to use the `CaseInsensitiveDict` from
       `request.structures`, because it turns out the lookup via that dict
       implementation is rather slow. So this version is somewhat of a
       trade-off, where I retain the same speed on lookups as a plain `dict`,
       but I also have a lower-cased key store, in case I ever need to use it.

    """
    __slots__ = ('_lower_store', )

    def __init__(self, data=None, **kwargs):
        super().__init__()
        self._lower_store = {}
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        # Store the lower-cased key for lookups via `get`. Also store the
        # actual key alongside the value.
        self._lower_store[key.lower()] = (key, value)

    def get_key(self, key) -> str:
        """Return the original cased key"""
        return self._lower_store[key.lower()][0]

    def get(self, key):
        """
        Do a case-insensitive lookup. This lower-cases `key` and looks up
        from the lower-cased key store.
        """
        try:
            return self.__getitem__(key)
        except KeyError:
            return self._lower_store[key.lower()][1]

    def __delitem__(self, key):
        lower_key = key.lower()
        actual_key, _ = self._lower_store[lower_key]

        del self[actual_key]
        del self._lower_store[lower_key]

    def lower_items(self):
        """Like iteritems(), but with all lowercase keys."""
        return (
            (lowerkey, keyval[1])
            for (lowerkey, keyval)
            in self._lower_store.items()
        )

    def __eq__(self, other):
        if isinstance(other, dict):
            other = DictWithLowerStore(other)
        else:
            return NotImplemented
        # Compare insensitively
        return dict(self.lower_items()) == dict(other.lower_items())

    def update(self, *args, **kwargs):
        if len(args) > 1:
            raise TypeError("update expected at most 1 arguments, got %d" % len(args))
        other = dict(*args, **kwargs)
        for key in other:
            self[key] = other[key]

    def copy(self):
        return DictWithLowerStore(self._lower_store.values())

    def __repr__(self):
        return str(dict(self.items()))



from __future__ import annotations

import json

from collections import Counter
from collections.abc import Mapping
from typing import TYPE_CHECKING, TypeVar

from dataclass_wizard.env import env_config


if TYPE_CHECKING:
    from dataclass_wizard._env import EnvInit
    T = TypeVar('T')


def envsafe(mapping: Mapping, *, dumps=json.dumps) -> dict[str, str]:
    out: dict[str, str] = {}
    for k, v in mapping.items():
        k = str(k)
        out[k] = v if isinstance(v, str) else dumps(v, separators=(",", ":"), sort_keys=True, default=str)
    return out


def from_env(cls: type[T], mapping: Mapping=None, env_cfg: EnvInit=None, **init_kwargs) -> T:
    env_map = envsafe(mapping or {})
    return cls(__env__=env_config(mapping=env_map, **(env_cfg or {})), **init_kwargs)


def assert_unordered_equal(a, b):
    if isinstance(a, dict) and isinstance(b, dict):
        assert a.keys() == b.keys()
        for k in a:
            assert_unordered_equal(a[k], b[k])
        return

    if isinstance(a, list) and isinstance(b, list):
        # compare as multisets, recursively
        def freeze(x):
            if isinstance(x, dict):
                return "dict", tuple(sorted((k, freeze(v)) for k, v in x.items()))
            if isinstance(x, list):
                # treat nested lists as unordered too
                return "list", tuple(sorted(freeze(v) for v in x))
            return "atom", x

        assert Counter(map(freeze, a)) == Counter(map(freeze, b))
        return

    assert a == b



"""
Wrapper utilities
"""
from typing import Callable


class FuncWrapper:
    """
    Wraps a callable `f` - which is occasionally useful, for example when
    defining functions as :class:`Enum` values. See below answer for more
    details.

    https://stackoverflow.com/a/40339397/10237506
    """
    __slots__ = ('f', )

    def __init__(self, f: Callable):
        self.f = f

    def __call__(self, *args, **kwargs):
        return self.f(*args, **kwargs)



import json
import logging
from dataclasses import MISSING, dataclass, fields
from typing import Callable

from .dumpers import asdict
from .lookups import Env, lookup_exact, clean
from ..abstractions import AbstractEnvWizard
from ..bases import AbstractEnvMeta
from ..bases_meta import BaseEnvWizardMeta, EnvMeta
from ..class_helper import (call_meta_initializer_if_needed, get_meta,
                            field_to_env_var, dataclass_field_to_json_field)
from ..decorators import cached_class_property
from ..enums import LetterCase
from ..environ.loaders import EnvLoader
from ..errors import ExtraData, MissingVars, ParseError, type_name
from ..loader_selection import get_loader
from ..log import enable_library_debug_logging
from ..models import Extras, JSONField
from ..type_def import ExplicitNull, JSONObject, dataclass_transform
from ..utils.function_builder import FunctionBuilder


_to_dataclass = dataclass(init=False)


@dataclass_transform(kw_only_default=True)
class EnvWizard(AbstractEnvWizard):
    """
    *Environment Wizard*

    A mixin class for parsing and managing environment variables in Python.

    ``EnvWizard`` makes it easy to map environment variables to Python attributes,
    handle defaults, and optionally load values from `.env` files.

    Quick Example::

        import os
        from pathlib import Path

        class MyConfig(EnvWizard):
            my_var: str
            my_optional_var: int = 42

        # Set environment variables
        os.environ["MY_VAR"] = "hello"

        # Load configuration from the environment
        config = MyConfig()
        print(config.my_var)  # Output: "hello"
        print(config.my_optional_var)  # Output: 42

        # Specify configuration explicitly
        config = MyConfig(my_var='world')
        print(config.my_var)  # Output: "world"
        print(config.my_optional_var)  # Output: 42

    Example with ``.env`` file::

        class MyConfigWithEnvFile(EnvWizard):
            class _(EnvWizard.Meta):
                env_file = True  # Defaults to loading from `.env`

            my_var: str
            my_optional_var: int = 42

        # Create an `.env` file in the current directory:
        # MY_VAR=world
        config = MyConfigWithEnvFile()
        print(config.my_var)  # Output: "world"
        print(config.my_optional_var)  # Output: 42

    Key Features:
        - Automatically maps environment variables to dataclass fields.
        - Supports default values for fields if environment variables are not set.
        - Optionally loads environment variables from `.env` files.
        - Supports prefixes for environment variables using ``_env_prefix`` or ``Meta.env_prefix``.
        - Supports loading secrets from directories using ``_secrets_dir`` or ``Meta.secrets_dir``.
        - Dynamic reloading with ``_reload`` to handle updated environment values.

    Initialization Options:
        The ``__init__`` method accepts additional parameters for flexibility:

        - ``_env_file`` (optional):
            Overrides the ``Meta.env_file`` value dynamically. Can be a file path,
            a sequence of file paths, or ``True`` to use the default `.env` file.
        - ``_reload`` (optional):
            Forces a reload of environment variables to bypass caching. Defaults to ``False``.
        - ``_env_prefix`` (optional):
            Dynamically overrides ``Meta.env_prefix``, applying a prefix to all environment
            variables. Defaults to ``None``.
        - ``_secrets_dir`` (optional):
            Overrides the ``Meta.secrets_dir`` value dynamically. Can be a directory path
            or a sequence of paths pointing to directories containing secret files.

    Meta Settings:
        These class-level attributes can be configured in a nested ``Meta`` class:

        - ``env_file``:
            The path(s) to `.env` files to load. If set to ``True``, defaults to `.env`.
        - ``env_prefix``:
            A prefix applied to all environment variables. Defaults to ``None``.
        - ``secrets_dir``:
            A path or sequence of paths to directories containing secret files. Defaults to ``None``.

    Attributes:
        Defined dynamically based on the dataclass fields in the derived class.
    """
    __slots__ = ()

    class Meta(BaseEnvWizardMeta):
        """
        Inner meta class that can be extended by sub-classes for additional
        customization with the environment load process.
        """
        __slots__ = ()

        # Class attribute to enable detection of the class type.
        __is_inner_meta__ = True

        def __init_subclass__(cls):
            # Set the `__init_subclass__` method here, so we can ensure it
            # doesn't run for the `EnvWizard.Meta` class.
            return cls._init_subclass()

    # noinspection PyMethodParameters,PyUnresolvedReferences
    @cached_class_property
    def __fields__(cls: type['E']):
        cls_fields = {}
        field_to_var = field_to_env_var(cls)

        for field in fields(cls):
            name = field.name
            cls_fields[name] = field

            if isinstance(field, JSONField):
                if not field.json.dump:
                    field_to_json_key = dataclass_field_to_json_field(cls)
                    field_to_json_key[name] = ExplicitNull

                keys = field.json.keys
                if keys:
                    # minor optimization: convert a one-element tuple of `str` to `str`
                    field_to_var[name] = keys[0] if len(keys) == 1 else keys

        return cls_fields

    to_dict = asdict

    def to_json(self, *,
                encoder = json.dumps,
                **encoder_kwargs):
        """
        Converts the `EnvWizard` subclass to a JSON `string` representation.
        """
        return encoder(asdict(self), **encoder_kwargs)

    def __init_subclass__(cls, *, reload_env=False, debug=False,
                          key_transform=LetterCase.NONE):

        if reload_env:  # reload cached var names from `os.environ` as needed.
            Env.reload()

        # apply the `@dataclass(init=False)` decorator to `cls`.
        _to_dataclass(cls)

        # set `key_transform_with_dump` for the class's Meta
        meta = EnvMeta(key_transform_with_dump=key_transform)

        if debug:
            # minimum logging level for logs by this library
            lvl = logging.DEBUG if isinstance(debug, bool) else debug
            # enable library logging
            enable_library_debug_logging(lvl)
            # set `debug_enabled` flag for the class's Meta
            meta.debug_enabled = lvl

        # Bind child class to DumpMeta with no key transformation.
        meta.bind_to(cls)

        # Calls the Meta initializer when inner :class:`Meta` is sub-classed.
        call_meta_initializer_if_needed(cls)

        # create and set methods such as `__init__()`.
        cls._create_methods()

    @classmethod
    def _create_methods(cls):
        """
        Generates methods such as the ``__init__()`` constructor method
        and ``dict()`` for the :class:`EnvWizard` subclass, vis-à-vis
        how the ``dataclasses`` module does it, with a few noticeable
        differences.
        """
        meta = get_meta(cls, base_cls=AbstractEnvMeta)
        cls_loader = get_loader(cls, base_cls=EnvLoader)

        # A cached mapping of each dataclass field name to its environment
        # variable name; useful so we don't need to do a case transformation
        # (via regex) each time.
        field_to_var = field_to_env_var(cls)

        # The function to case-transform and lookup variables defined in the
        # environment.
        get_env: 'Callable[[str], str | None]' = meta.key_lookup_with_load

        # noinspection PyArgumentList
        extras = Extras(config=None)

        cls_fields = cls.__fields__
        field_names = frozenset(cls_fields)

        _meta_env_file = meta.env_file

        _locals = {'Env': Env,
                   'ParseError': ParseError,
                   'field_names': field_names,
                   'get_env': get_env,
                   'lookup_exact': lookup_exact}

        _globals = {'MissingVars': MissingVars,
                    'add': _add_missing_var,
                    'cls': cls,
                    'fields_ordered': cls_fields.keys(),
                    'handle_err': _handle_parse_error,
                    'MISSING': MISSING,
                    }

        if meta.secrets_dir is None:
            _secrets_dir_value = 'None'
        else:
            _locals['_secrets_dir_value'] = meta.secrets_dir
            _secrets_dir_value = '_secrets_dir_value'

        # parameters to the `__init__()` method.
        init_params = ['self',
                       '_env_file=None',
                       '_reload=False',
                       f'_env_prefix={meta.env_prefix!r}',
                       f'_secrets_dir={_secrets_dir_value}',
        ]

        fn_gen = FunctionBuilder()

        with fn_gen.function('__init__', init_params, None, _locals):

            # reload cached var names from `os.environ` as needed.
            with fn_gen.if_('_reload'):
                fn_gen.add_line('Env.reload()')
            with fn_gen.else_():
                fn_gen.add_line('Env.load_environ()')

            with fn_gen.if_('_secrets_dir'):
                fn_gen.add_line('Env.update_with_secret_values(_secrets_dir)')

            # update environment with values in the "dot env" files as needed.
            if _meta_env_file:
                fn = fn_gen.elif_
                _globals['_dotenv_values'] = Env.dotenv_values(_meta_env_file)
                with fn_gen.if_('_env_file is None'):
                    fn_gen.add_line('Env.update_with_dotenv(dotenv_values=_dotenv_values)')
            else:
                fn = fn_gen.if_
            with fn('_env_file'):
                fn_gen.add_line('Env.update_with_dotenv(_env_file)')

            # iterate over the dataclass fields and (attempt to) resolve
            # each one.
            fn_gen.add_line('_vars = []')

            if field_names:

                with fn_gen.try_():

                    for name, f in cls_fields.items():
                        type_field = f'_tp_{name}'
                        tp = _globals[type_field] = f.type

                        init_params.append(f'{name}:{type_field}=MISSING')

                        # retrieve value (if it exists) for the environment variable

                        env_var = var_name = field_to_var.get(name)
                        if env_var:
                            part = f'({name} := lookup_exact(_var_name))'
                        else:
                            var_name = name
                            part = f'({name} := get_env(_var_name))'

                        fn_gen.add_line(f'_name={name!r}; _env_var={env_var!r}; _var_name=f"{{_env_prefix}}{var_name}" if _env_prefix else {var_name!r}')

                        with fn_gen.if_(f'{name} is not MISSING or {part} is not MISSING'):
                            parser_name = f'_parser_{name}'
                            _globals[parser_name] = getattr(p := cls_loader.get_parser_for_annotation(
                                tp, cls, extras), '__call__', p)
                            fn_gen.add_line(f'self.{name} = {parser_name}({name})')
                        # this `else` block means that a value was not received for the
                        # field, either via keyword arguments or Environment.
                        with fn_gen.else_():
                            # check if the field defines a `default` or `default_factory`
                            # value; note this is similar to how `dataclasses` does it.
                            default_name = f'_dflt_{name}'
                            if f.default is not MISSING:
                                _globals[default_name] = f.default
                                fn_gen.add_line(f'self.{name} = {default_name}')
                            elif f.default_factory is not MISSING:
                                _globals[default_name] = f.default_factory
                                fn_gen.add_line(f'self.{name} = {default_name}()')
                            else:
                                fn_gen.add_line(f'add(_vars, _name, _env_prefix, _env_var, {type_field})')

                with fn_gen.except_(ParseError, 'e'):
                    fn_gen.add_line('handle_err(e, cls, _name, _env_prefix, _env_var)')

            # check for any required fields with missing values
            with fn_gen.if_('_vars'):
                fn_gen.add_line('raise MissingVars(cls, _vars) from None')

            # if keyword arguments are passed in, confirm that all there
            # aren't any "extra" keyword arguments
            # if _extra is not Extra.IGNORE:
                # with fn_gen.if_('has_kwargs'):
                #     # get a list of keyword arguments that don't map to any fields
                #     fn_gen.add_line('extra_kwargs = set(init_kwargs) - field_names')
                #     with fn_gen.if_('extra_kwargs'):
                #         # the default behavior is "DENY", so an error will be raised here.
                #         if _extra is None or _extra is Extra.DENY:
                #             _globals['ExtraData'] = ExtraData
                #             fn_gen.add_line('raise ExtraData(cls, extra_kwargs, list(fields_ordered)) from None')
                #         else:  # Extra.ALLOW
                #             # else, if we want to "ALLOW" extra keyword arguments, we need to
                #             # store those attributes in the instance.
                #             with fn_gen.for_('attr in extra_kwargs'):
                #                 fn_gen.add_line('setattr(self, attr, init_kwargs[attr])')

        with fn_gen.function('dict', ['self'], JSONObject, _locals):
            parts = ','.join([f'{name!r}:self.{name}' for name, f in cls.__fields__.items()])
            fn_gen.add_line(f'return {{{parts}}}')

        functions = fn_gen.create_functions(_globals)

        # set the `__init__()` method.
        cls.__init__ = functions['__init__']
        # set the `dict()` method.
        cls.dict = functions['dict']


def _add_missing_var(missing_vars, name, env_prefix, var_name, tp):

    var_name = _get_var_name(name, env_prefix, var_name)
    tn = type_name(tp)
    # noinspection PyBroadException
    try:
        suggested = tp()
    except Exception:
        suggested = None

    missing_vars.append((name, var_name, tn, suggested))


def _handle_parse_error(e, cls, name, env_prefix, var_name):

    # We run into a parsing error while loading the field
    # value; Add additional info on the Exception object
    # before re-raising it.
    e.class_name = cls
    e.field_name = name
    e.kwargs['env_variable'] = _get_var_name(name, env_prefix, var_name)

    raise


def _get_var_name(name, env_prefix, var_name):

    if var_name is None:
        env_var = f'{env_prefix}{name}' if env_prefix else name
        var_name = Env.cleaned_to_env.get(clean(env_var), env_var)

    elif env_prefix:
        var_name = f'{env_prefix}{var_name}'

    return var_name



# noinspection PyProtectedMember
from collections.abc import Mapping, Sequence
from dataclasses import _MISSING_TYPE, MISSING
from dataclasses import Field as _Field
from typing import Any, Literal, TypeAlias, overload

from ._type_def import DefFactory, T
from .conditions import Condition
from .utils._object_path import PathType

# Define a simple type (alias) for the `CatchAll` field
CatchAll: TypeAlias = Mapping | None

# noinspection PyPep8Naming
def AliasPath(*all: PathType | str,
              load: PathType | str | None = None,
              dump: PathType | str | None = None,
              skip: bool = False,
              default: Any = MISSING,
              default_factory: DefFactory[T] | Literal[_MISSING_TYPE.MISSING] = MISSING,
              init: bool = True,
              repr: bool = True,
              hash: bool | None = None,
              compare: bool = True,
              metadata: Mapping[Any, Any] | None = None,
              kw_only: bool = False) -> Field:
    """
    Creates a dataclass field mapped to one or more nested JSON paths.

    This function acts as an alias for ``dataclasses.field(...)``, with additional
    functionality to associate a field with one or more nested JSON paths,
    including complex or deeply nested structures.

    The mapping is case-sensitive, meaning that JSON keys must match exactly
    (e.g., "myField" will not match "myfield"). Nested paths can include dot
    notations or bracketed syntax for accessing specific indices or keys.

    :param all: One or more nested JSON paths to associate with
        the dataclass field (e.g., ``a.b.c`` or ``a["nested"]["key"]``).
    :type all: PathType | str
    :param load: Path(s) to use for deserialization. Defaults to ``all`` if not specified.
    :type load: PathType | str | None
    :param dump: Path(s) to use for serialization. Defaults to ``all`` if not specified.
    :type dump: PathType | str | None
    :param skip: If True, the field is excluded during serialization. Defaults to False.
    :type skip: bool
    :param default: Default value for the field. Cannot be used with ``default_factory``.
    :type default: Any
    :param default_factory: A callable to generate the default value. Cannot be used with ``default``.
    :type default_factory: Callable[[], Any]
    :param init: Whether the field is included in the generated ``__init__`` method. Defaults to True.
    :type init: bool
    :param repr: Whether the field appears in the ``__repr__`` output. Defaults to True.
    :type repr: bool
    :param hash: Whether the field is included in the ``__hash__`` method. Defaults to None.
    :type hash: bool
    :param compare: Whether the field is included in comparison methods. Defaults to True.
    :type compare: bool
    :param metadata: Additional metadata for the field. Defaults to None.
    :type metadata: dict
    :param kw_only: If True, the field is keyword-only. Defaults to False.
    :type kw_only: bool
    :return: A dataclass field with additional mapping to one or more nested JSON paths.
    :rtype: Field

    **Examples**

    **Example 1** -- Mapping multiple nested paths to a field::

        from dataclasses import dataclass

        from dataclass_wizard import AliasPath, fromdict

        @dataclass
        class Example:
            my_str: str = AliasPath('a.b.c.1', 'x.y["-1"].z', default="default_value")

        # Maps nested paths ('a', 'b', 'c', 1) and ('x', 'y', '-1', 'z')
        # to the `my_str` attribute. '-1' is treated as a literal string key,
        # not an index, for the second path.

        print(fromdict(Example, {'x': {'y': {'-1': {'z': 'some_value'}}}}))
        #> Example(my_str='some_value')

    **Example 2** -- Using Annotated::

        from dataclasses import dataclass
        from typing import Annotated

        from dataclass_wizard import AliasPath, JSONWizard

        @dataclass
        class Example(JSONWizard):
            my_str: Annotated[str, AliasPath('my."7".nested.path.-321')]


        ex = Example.from_dict({'my': {'7': {'nested': {'path': {-321: 'Test'}}}}})
        print(ex)  #> Example(my_str='Test')
    """


# noinspection PyPep8Naming
def Alias(*all: str,
          load: str | Sequence[str] | None = None,
          dump: str | None = None,
          env: str | Sequence[str] | None = None,
          skip: bool = False,
          default=MISSING,
          default_factory: DefFactory[T] | Literal[_MISSING_TYPE.MISSING] = MISSING,
          init=True, repr=True,
          hash=None, compare=True, metadata=None, kw_only=False):
    """
    Maps one or more JSON key names to a dataclass field.

    This function acts as an alias for ``dataclasses.field(...)``, with additional
    support for associating a field with one or more JSON keys. It customizes
    serialization and deserialization behavior, including handling keys with
    varying cases or alternative names.

    The mapping is case-sensitive; JSON keys must match exactly (e.g., ``myField``
    will not match ``myfield``). If multiple keys are provided, the first one
    is used as the default for serialization.

    :param all: One or more JSON key names to associate with the dataclass field.
    :type all: str
    :param load: Key(s) to use for deserialization. Defaults to ``all`` if not specified.
    :type load: str | Sequence[str] | None
    :param dump: Key to use for serialization. Defaults to the first key in ``all``.
    :type dump: str | None
    :param env: Environment variable(s) to use for deserialization.
    :type env: str | Sequence[str] | None
    :param skip: If ``True``, the field is excluded during serialization. Defaults to ``False``.
    :type skip: bool
    :param default: Default value for the field. Cannot be used with ``default_factory``.
    :type default: Any
    :param default_factory: Callable to generate the default value. Cannot be used with ``default``.
    :type default_factory: Callable[[], Any]
    :param init: Whether the field is included in the generated ``__init__`` method. Defaults to ``True``.
    :type init: bool
    :param repr: Whether the field appears in the ``__repr__`` output. Defaults to ``True``.
    :type repr: bool
    :param hash: Whether the field is included in the ``__hash__`` method. Defaults to ``None``.
    :type hash: bool
    :param compare: Whether the field is included in comparison methods. Defaults to ``True``.
    :type compare: bool
    :param metadata: Additional metadata for the field. Defaults to ``None``.
    :type metadata: dict
    :param kw_only: If ``True``, the field is keyword-only. Defaults to ``False``.
    :type kw_only: bool
    :return: A dataclass field with additional mappings to one or more JSON keys.
    :rtype: Field

    **Examples**

    **Example 1** -- Mapping multiple key names to a field::

        from dataclasses import dataclass

        from dataclass_wizard import Alias, fromdict

        @dataclass
        class Example:
            my_field: str = Alias('key1', 'key2', default="default_value")

        print(fromdict(Example, {'key2': 'a value!'}))
        #> Example(my_field='a value!')

    **Example 2** -- Skipping a field during serialization::

        from dataclasses import dataclass

        from dataclass_wizard import Alias, JSONWizard

        @dataclass
        class Example(JSONWizard):
            my_field: str = Alias('key', skip=True)

        ex = Example.from_dict({'key': 'some value'})
        print(ex)                  #> Example(my_field='a value!')
        assert ex.to_dict() == {}  #> True
    """


# noinspection PyPep8Naming
def Env(*load: str,
        default=MISSING,
        default_factory: DefFactory[T] | Literal[_MISSING_TYPE.MISSING] = MISSING,
        init=True, repr=True,
        hash=None, compare=True, metadata=None, kw_only=False):
    """
    Maps one or more Environment Variable names to a dataclass field.

    This function acts as an alias for ``dataclasses.field(...)``, with additional
    support for associating a field with one or more env vars. It customizes
    serialization and deserialization behavior, including handling env vars with
    varying cases or alternative names.

    The mapping is case-sensitive; env vars must match exactly (e.g., ``myField``
    will not match ``myfield``).

    :param load: Env vars(s) to use for deserialization.
    :type load: str
    :param default: Default value for the field. Cannot be used with ``default_factory``.
    :type default: Any
    :param default_factory: Callable to generate the default value. Cannot be used with ``default``.
    :type default_factory: Callable[[], Any]
    :param init: Whether the field is included in the generated ``__init__`` method. Defaults to ``True``.
    :type init: bool
    :param repr: Whether the field appears in the ``__repr__`` output. Defaults to ``True``.
    :type repr: bool
    :param hash: Whether the field is included in the ``__hash__`` method. Defaults to ``None``.
    :type hash: bool
    :param compare: Whether the field is included in comparison methods. Defaults to ``True``.
    :type compare: bool
    :param metadata: Additional metadata for the field. Defaults to ``None``.
    :type metadata: dict
    :param kw_only: If ``True``, the field is keyword-only. Defaults to ``False``.
    :type kw_only: bool
    :return: A dataclass field with additional mappings to one or more JSON keys.
    :rtype: Field

    **Examples**

    **Example 1** -- Mapping multiple key names to a field::

        from dataclasses import dataclass

        from dataclass_wizard import Alias, fromdict

        @dataclass
        class Example:
            my_field: str = Alias('key1', 'key2', default="default_value")

        print(fromdict(Example, {'key2': 'a value!'}))
        #> Example(my_field='a value!')

    **Example 2** -- Skipping a field during serialization::

        from dataclasses import dataclass

        from dataclass_wizard import Alias, JSONWizard

        @dataclass
        class Example(JSONWizard):
            my_field: str = Alias('key', skip=True)

        ex = Example.from_dict({'key': 'some value'})
        print(ex)                  #> Example(my_field='a value!')
        assert ex.to_dict() == {}  #> True
    """


def skip_if_field(condition: Condition, *,
                  default=MISSING,
                  default_factory: DefFactory[T] | Literal[_MISSING_TYPE.MISSING] = MISSING,
                  init=True, repr=True,
                  hash=None, compare=True, metadata=None,
                  kw_only: bool | Literal[_MISSING_TYPE.MISSING] = MISSING):
    """
    Defines a dataclass field with a ``SkipIf`` condition.

    This function is a shortcut for ``dataclasses.field(...)``,
    adding metadata to specify a condition. If the condition
    evaluates to ``True``, the field is skipped during
    JSON serialization.

    Arguments:
        condition (Condition): The condition, if true skips serializing the field.
        default (Any): The default value for the field. Mutually exclusive with `default_factory`.
        default_factory (Callable[[], Any]): A callable to generate the default value.
                                             Mutually exclusive with `default`.
        init (bool): Include the field in the generated `__init__` method. Defaults to True.
        repr (bool): Include the field in the `__repr__` output. Defaults to True.
        hash (bool): Include the field in the `__hash__` method. Defaults to None.
        compare (bool): Include the field in comparison methods. Defaults to True.
        metadata (dict): Metadata to associate with the field. Defaults to None.
        kw_only (bool): If true, the field will become a keyword-only parameter to __init__().
    Returns:
        Field: A dataclass field with correct metadata set.

    Example:
        >>> from dataclasses import dataclass
        >>> from dataclass_wizard.conditions import IS_NOT
        >>> @dataclass
        >>> class Example:
        >>>     my_str: str = skip_if_field(IS_NOT(True))
        >>> # Creates a condition which skips serializing `my_str`
        >>> # if its value `is not True`.
    """


class Field(_Field):
    """
    Alias to a :class:`dataclasses.Field`, but one which also represents a
    mapping of one or more JSON key names to a dataclass field.

    See the docs on the :func:`Alias` and :func:`AliasPath` for more info.
    """
    __slots__ = ('load_alias',
                 'dump_alias',
                 'env_vars',
                 'skip',
                 'path')

    load_alias: str | None
    dump_alias: str | None
    env_vars: str | None
    skip: bool
    path: PathType | None

    # In Python 3.14, dataclasses adds a new parameter to the :class:`Field`
    # constructor: `doc`
    #
    # Ref: https://docs.python.org/3.14/library/dataclasses.html#dataclasses.field
    @overload
    def __init__(self,
                 load_alias: str | None,
                 dump_alias: str | None,
                 env_vars: str | None,
                 skip: bool,
                 path: PathType | None,
                 default, default_factory, init, repr, hash, compare,
                 metadata, kw_only, doc):
        ...

    # In Python 3.10, dataclasses adds a new parameter to the :class:`Field`
    # constructor: `kw_only`
    #
    # Ref: https://docs.python.org/3.10/library/dataclasses.html#dataclasses.dataclass
    @overload
    def __init__(self,
                 load_alias: str | None,
                 dump_alias: str | None,
                 env_vars: str | None,
                 skip: bool,
                 path: PathType | None,
                 default, default_factory, init, repr, hash, compare,
                 metadata, kw_only):
        ...

    @overload
    def __init__(self,
                 load_alias: str | None,
                 dump_alias: str | None,
                 env_vars: str | None,
                 skip: bool,
                 path: PathType | None,
                 default, default_factory, init, repr, hash, compare,
                 metadata):
        ...



__all__ = [
    'snake',
    'does_not_raise',
    'data_file_path',
]

import sys
# Ref: https://docs.pytest.org/en/6.2.x/example/parametrize.html#parametrizing-conditional-raising
from contextlib import nullcontext as does_not_raise
from logging import getLogger
from pathlib import Path

import pytest

from dataclass_wizard.constants import PACKAGE_NAME
from dataclass_wizard.utils._string_case import to_snake_case

from ._typing import PY312_OR_ABOVE


# Directory for test files
TEST_DATA_DIR = Path(__file__).resolve().parent / 'testdata'


collect_ignore = []

# Ignore test files if the Python version is below 3.12
if not PY312_OR_ABOVE:
    print("Python version is below 3.12. Ignoring test files.")
    collect_ignore.append(
        str(Path(__file__).parent / "unit" / "test_union_as_type_alias_recursive.py")
    )


@pytest.fixture
def restore_logger():
    logger = getLogger(PACKAGE_NAME)

    old_level = logger.level
    old_propagate = logger.propagate
    old_handlers = list(logger.handlers)

    yield logger

    # remove any new handlers
    for h in list(logger.handlers):
        if h not in old_handlers:
            logger.removeHandler(h)
            try:
                h.close()
            except Exception:
                pass

    # restore original handlers (in case code removed/changed them)
    logger.handlers[:] = old_handlers
    logger.setLevel(old_level)
    logger.propagate = old_propagate


def data_file_path(name: str) -> str:
    """Returns the full path to a test file."""
    return str((TEST_DATA_DIR / name).absolute())


def snake(d):
    """
    Helper function to snake-case all keys in a dictionary `d`.

    Useful for `v1`, which by default requires a 1:1 mapping of
    JSON key to dataclass field.
    """
    return {to_snake_case(k): v for k, v in d.items()}



from collections.abc import Mapping, Sequence
from typing import Callable
from weakref import WeakKeyDictionary, WeakSet

from ._abstractions import (
    AbstractDumperGenerator,
    AbstractLoaderGenerator,
    E,
    W,
)
from ._type_def import T
from .conditions import Condition
from .constants import PACKAGE_NAME
from .utils._object_path import PathType

# A mapping of dataclass to its loader.
CLASS_TO_LOADER: WeakKeyDictionary[type, type[AbstractLoaderGenerator]]

# A mapping of dataclass to its dumper.
CLASS_TO_DUMPER: WeakKeyDictionary[type, type[AbstractDumperGenerator]]

# We use a sentinel mapping to confirm if we need to set up the load
# config for a dataclass on an initial run.
IS_CONFIG_SETUP: WeakSet[type]

# A cached mapping, per dataclass, of instance field name to JSON path
DATACLASS_FIELD_TO_ALIAS_PATH_FOR_LOAD: WeakKeyDictionary[
    type, dict[str, Sequence[PathType]]]

# Dump: A cached mapping, per dataclass, of instance field name to alias path
DATACLASS_FIELD_TO_ALIAS_PATH_FOR_DUMP: WeakKeyDictionary[
    type, dict[str, Sequence[PathType]]]

# A cached mapping, per dataclass, of instance field name to alias
DATACLASS_FIELD_TO_ALIAS_FOR_LOAD: WeakKeyDictionary[
    type, dict[str, Sequence[str]]]

# A cached mapping, per dataclass, of instance field name to env var
DATACLASS_FIELD_TO_ENV_FOR_LOAD: WeakKeyDictionary[
    type, dict[str, Sequence[str]]]

# A cached mapping, per dataclass, of instance field name to alias
DATACLASS_FIELD_TO_ALIAS_FOR_DUMP: WeakKeyDictionary[
    type, dict[str, str]]

# A cached mapping, per dataclass, of instance field name to `SkipIf` condition
DATACLASS_FIELD_TO_SKIP_IF: WeakKeyDictionary[
    type, dict[str, Condition]]

# Cache: owner class -> its `Meta` inner class (only present when subclassed)
META_INITIALIZER: dict[str, Callable[[type[W]], None]] = {}

def set_class_loader(
        cls_to_loader: Mapping[type, type[AbstractLoaderGenerator]],
        class_or_instance: type[T] | T,
        loader: type[AbstractLoaderGenerator]):
    """
    Set (and return) the loader for a dataclass.
    """

def set_class_dumper(
        cls_to_dumper: Mapping[type, type[AbstractDumperGenerator]],
        class_or_instance: type[T] | T,
        dumper: type[AbstractDumperGenerator]):
    """
    Set (and return) the dumper for a dataclass.
    """

def dataclass_field_to_skip_if(cls: type) -> dict[str, Condition]:
    """
    Returns a mapping of dataclass field to SkipIf condition.
    """

def resolve_dataclass_field_to_alias_for_dump(
    cls: type) -> dict[str, Sequence[str]]: ...
def resolve_dataclass_field_to_alias_for_load(
    cls: type) -> dict[str, Sequence[str]]: ...
def resolve_dataclass_field_to_env_for_load(
    cls: type) -> dict[str, Sequence[str]]: ...

def setup_config_for_cls(cls: type):
    """
    This function processes a class `cls` on an initial run, and sets up the
    load process for `cls` by iterating over each dataclass field. For each
    field, it performs the following tasks:

        * Check if the field's annotation is of type ``Annotated``. If so,
          we iterate over each ``Annotated`` argument and find any special
          :class:`JSON` objects (this can also be set via the helper function
          ``json_key``). Assuming we find it, the class-specific mapping of
          dataclass field name to JSON key is then updated with the input
          passed in to this object.

        * Check if the field type is a :class:`JSONField` object (this can
          also be set by the helper function ``json_field``). Assuming this is
          the case, the class-specific mapping of dataclass field name to
          JSON key is then updated with the input passed in to
          the :class:`JSON` attribute.
    """

def call_meta_initializer_if_needed(cls: type[W | E],
                                    package_name=PACKAGE_NAME) -> None:
    """
    Calls the Meta initializer when the inner :class:`Meta` is sub-classed.
    """



from collections import namedtuple
from dataclasses import dataclass
from datetime import datetime, date, timezone
from typing import Tuple, NamedTuple, List

import pytest

from dataclass_wizard import DataclassWizard, EnvWizard

from ..utils_env import from_env


def test_load_to_bytes():
    class E(EnvWizard):
        b: bytes

    e = E(b='testing 123')
    assert e.b == b'testing 123'


@pytest.mark.parametrize(
    'input,expected',
    [
        ('testing 123', bytearray(b'testing 123')),
        (b'test', bytearray(b'test')),
        ([1, 2, 3], bytearray([1, 2, 3]))
    ]
)
def test_load_to_bytearray(input, expected):
    class MyClass(EnvWizard):
        my_btarr: bytearray

    assert MyClass(my_btarr=input).my_btarr == expected


def test_load_to_tuple_and_named_tuple():

    class MyNT(NamedTuple):
        my_float: float
        my_str: str

    untyped_tup = namedtuple('untyped_tup', ('a', 'b', 'c'))

    class MyClass(EnvWizard):
        my_tup: Tuple[int, ...]
        my_nt: MyNT
        my_untyped_nt: untyped_tup

    env = {'MY_TUP': '1,2,3',
           'MY_NT': '[1.23, "string"]',
           'my_untyped_nt': 'hello ,  world, 123'}

    c = from_env(MyClass, env)

    assert c.raw_dict() == {
        'my_nt': MyNT(my_float=1.23, my_str='string'),
        'my_tup': (1, 2, 3),
        'my_untyped_nt': untyped_tup(a='hello', b='world', c='123'),
    }

    assert c.to_dict() == {'my_nt': [1.23, 'string'],
                           'my_tup': [1, 2, 3],
                           'my_untyped_nt': ['hello', 'world', '123']}


def test_load_to_dataclass():
    """When an `EnvWizard` subclass has a nested dataclass schema."""

    @dataclass
    class Inner1:
        my_bool: bool
        my_string: str

    class Inner2(DataclassWizard, load_case='AUTO'):
        answer_to_life: int
        my_list: List[str]

    class MyClass(EnvWizard):

        inner_cls_1: Inner1
        inner_cls_2: Inner2

    env = {'inner_cls_1': 'my_bool=false, my_string=test',
           'inner_cls_2': '{"answerToLife": "42", "MyList": "testing, 123 , hello!"}'}

    c = from_env(MyClass, env)
    # print(c)

    assert c.raw_dict() == {
        'inner_cls_1': Inner1(my_bool=False,
                              my_string='test'),
        'inner_cls_2': Inner2(answer_to_life=42,
                              my_list=['testing', '123', 'hello!']),
    }

    assert c.to_dict() == {
        'inner_cls_1': {'my_bool': False,
                        'my_string': 'test'},
        'inner_cls_2': {'answer_to_life': 42,
                        'my_list': ['testing', '123', 'hello!']}
    }


@pytest.mark.parametrize(
    'input,expected',
    [
        ('2021-11-28T17:35:55', datetime(2021, 11, 28, 17, 35, 55)),
        (1577952245, datetime(2020, 1, 2, 8, 4, 5, tzinfo=timezone.utc)),
        (datetime.min, datetime.min)
    ]
)
def test_load_to_datetime(input, expected):
    class MyClass(EnvWizard):
        my_dt: datetime

    assert MyClass(my_dt=input).my_dt == expected


@pytest.mark.parametrize(
    'input,expected',
    [
        ('2021-11-28', date(2021, 11, 28)),
        (1577952245, date(2020, 1, 2)),
        (date.min, date.min)
    ]
)
def test_load_to_date(input, expected):
    class MyClass(EnvWizard):
        my_date: date

    assert MyClass(my_date=input).my_date == expected



import logging
from abc import ABC
from base64 import b64decode
from collections import deque, defaultdict
from dataclasses import dataclass, field
from datetime import datetime, timedelta, timezone, date
from typing import (Set, FrozenSet, Optional, Union, List,
                    DefaultDict, Annotated, Literal)
from uuid import UUID

import pytest

from dataclass_wizard import *
from dataclass_wizard._meta_cache import get_meta
from dataclass_wizard.constants import TAG
from dataclass_wizard.errors import ParseError
from dataclass_wizard.enums import KeyAction
from tests.unit.conftest import *
from tests._typing import *

log = logging.getLogger(__name__)


def test_asdict_and_fromdict():
    """
    Confirm that Meta settings for both `fromdict` and `asdict` are merged
    as expected.
    """

    @dataclass
    class MyClass:
        my_bool: Optional[bool]
        myStrOrInt: Union[str, int]

    d = {'myBoolean': 'tRuE', 'myStrOrInt': 123}

    LoadMeta(
        case='CAMEL',
        on_unknown_key='RAISE',
        field_to_alias={'my_bool': 'myBoolean'},
    ).bind_to(MyClass)

    # Keep same dump output as before: `myBoolean` for my_bool + snake for the rest.
    DumpMeta(
        case='SNAKE',
        field_to_alias={'myStrOrInt': 'My String-Or-Num'},
    ).bind_to(MyClass)

    meta = get_meta(MyClass)

    # The library normalizes these internally; accept common representations.
    assert meta.case is None

    assert str(meta.load_case).upper() in ('CAMEL', 'C')
    assert str(meta.dump_case).upper() in ('SNAKE', 'S')
    assert meta.on_unknown_key is KeyAction.RAISE
    assert meta.field_to_alias_load == {'my_bool': 'myBoolean'}
    assert meta.field_to_alias_dump == {'myStrOrInt': 'My String-Or-Num'}

    c = fromdict(MyClass, d)

    assert c.my_bool is True
    assert isinstance(c.myStrOrInt, int)
    assert c.myStrOrInt == 123

    new_dict = asdict(c)
    assert new_dict == {'my_bool': True, 'My String-Or-Num': 123}


def test_asdict_with_nested_dataclass():
    """Confirm that `asdict` works for nested dataclasses as well."""

    @dataclass
    class Container:
        id: int
        submittedDate: date
        submittedDt: datetime
        myElements: List['MyElement']

    @dataclass
    class MyElement:
        order_index: Optional[int]
        status_code: Union[int, str]

    submitted_date = date(2019, 11, 30)
    naive_dt = datetime(2021, 1, 1, 5)
    elements = [MyElement(111, '200'), MyElement(222, 404)]

    # Fix so the forward reference works (since the class definition is inside
    # the test case)
    globals().update(locals())

    DumpMeta(
        case='SNAKE',
        dump_date_time_as='TIMESTAMP',
        assume_naive_datetime_tz=timezone.utc,
    ).bind_to(Container)

    # Case 1: naive dt -> assumed UTC -> timestamp
    c1 = Container(123, submitted_date, naive_dt, myElements=elements)
    d1 = asdict(c1)

    expected1 = {
        "id": 123,
        "submitted_date": round(datetime(2019, 11, 30, tzinfo=timezone.utc).timestamp()),
        "submitted_dt": round(naive_dt.replace(tzinfo=timezone.utc).timestamp()),
        "my_elements": [
            {"order_index": 111, "status_code": "200"},
            {"order_index": 222, "status_code": 404},
        ],
    }
    assert d1 == expected1

    # Case 2: aware dt (fixed offset "EST") -> convert to UTC -> timestamp
    est_fixed = timezone(timedelta(hours=-5))
    aware_dt = naive_dt.replace(tzinfo=est_fixed)

    c2 = Container(123, submitted_date, aware_dt, myElements=elements)
    d2 = asdict(c2)

    expected2 = dict(expected1)
    expected2["submitted_dt"] = round(aware_dt.timestamp())
    assert d2 == expected2


def test_tag_field_is_used_in_dump_process():
    """
    Confirm that the `_TAG` field appears in the serialized JSON or dict
    object (even for nested dataclasses) when a value is set in the
    `Meta` config for a JSONWizard sub-class.
    """

    @dataclass
    class Data(ABC):
        """ base class for a Member """
        number: float

    class DataA(Data):
        """ A type of Data"""
        pass

    class DataB(Data, JSONWizard):
        """ Another type of Data """

        class _(JSONWizard.Meta):
            """
            This defines a custom tag that shows up in de-serialized
            dictionary object.
            """
            tag = 'B'

    @dataclass
    class Container(JSONWizard):
        """ container holds a subclass of Data """

        class _(JSONWizard.Meta):
            tag = 'CONTAINER'

        data: Union[DataA, DataB]

    data_a = DataA(number=1.0)
    data_b = DataB(number=1.0)

    container = Container(data=data_a)
    d1 = container.to_dict()

    # TODO: Right now `tag` is only populated for dataclasses in `Union`,
    #  but I don't think it's a big issue.

    expected = {
        # TAG: 'CONTAINER',
        'data': {'number': 1.0}
    }
    assert d1 == expected

    container = Container(data=data_b)
    d2 = container.to_dict()

    expected = {
        # TAG: 'CONTAINER',
        'data': {
            TAG: 'B',
            'number': 1.0
        }
    }
    assert d2 == expected


def test_to_dict_key_transform_with_json_field():
    """
    Specifying a custom mapping of JSON key to dataclass field.

    v1: use Alias(...) instead of json_field/json_key.
    """

    @dataclass
    class MyClass(JSONWizard):

        my_str: str = Alias('myCustomStr')
        my_bool: bool = Alias('my_json_bool', 'myTestBool')

    value = 'Testing'
    expected = {'myCustomStr': value, 'my_json_bool': True}

    c = MyClass(my_str=value, my_bool=True)

    result = c.to_dict()
    log.debug('Parsed object: %r', result)

    assert result == expected


def test_to_dict_key_transform_with_json_key():
    """
    Specifying a custom mapping of JSON key to dataclass field.

    v1: use Annotated[..., Alias(...)].
    """

    @dataclass
    class MyClass(JSONWizard):
        my_str: Annotated[str, Alias('myCustomStr')]
        my_bool: Annotated[bool, Alias('my_json_bool', 'myTestBool')]

    value = 'Testing'
    expected = {'myCustomStr': value, 'my_json_bool': True}

    c = MyClass(my_str=value, my_bool=True)

    result = c.to_dict()
    log.debug('Parsed object: %r', result)

    result = c.to_dict()
    log.debug('Parsed object: %r', result)

    assert result == expected


def test_to_dict_with_skip_defaults():
    """
    When `skip_defaults` is enabled in the class Meta, fields with default
    values should be excluded from the serialization process.
    """

    @dataclass
    class MyClass(JSONWizard):
        class _(JSONWizard.Meta):
            dump_case = 'C'
            skip_defaults = True

        my_str: str
        other_str: str = 'any value'
        optional_str: str = None
        my_list: List[str] = field(default_factory=list)
        my_dict: DefaultDict[str, List[float]] = field(
            default_factory=lambda: defaultdict(list))

    c = MyClass('abc')
    log.debug('Instance: %r', c)

    out_dict = c.to_dict()
    assert out_dict == {'myStr': 'abc'}


def test_to_dict_with_excluded_fields():
    """
    Excluding dataclass fields from the serialization process works
    as expected.
    """

    @dataclass
    class MyClass(JSONWizard):
        my_str: str
        # v1: map load alias + disable dump
        other_str: Annotated[str, Alias(load='AnotherStr', skip=True)]
        my_bool: bool = Alias(load='TestBool', skip=True)
        my_int: int = 3

    data = {'my_str': 'my string',
            'AnotherStr': 'testing 123',
            'TestBool': True}

    c = MyClass.from_dict(data)
    log.debug('Instance: %r', c)

    additional_exclude = ('my_int', )

    out_dict = c.to_dict(exclude=additional_exclude)
    assert out_dict == {'my_str': 'my string'}


@pytest.mark.parametrize(
    'input,expected,expectation',
    [
        ({1, 2, 3}, [1, 2, 3], does_not_raise()),
        ((3.22, 2.11, 1.22), [3.22, 2.11, 1.22], does_not_raise()),
    ]
)
def test_set(input, expected, expectation):

    @dataclass
    class MyClass(JSONWizard):
        num_set: Set[int]
        any_set: set

    expected = sorted(expected)

    input_set = set(input)
    c = MyClass(num_set=input_set, any_set=input_set)

    with expectation:
        result = c.to_dict()
        log.debug('Parsed object: %r', result)

        assert all(key in result for key in ('num_set', 'any_set'))
        assert isinstance(result['num_set'], (list, tuple))
        assert isinstance(result['any_set'], (list, tuple))

        assert sorted(result['num_set']) == expected
        assert sorted(result['any_set']) == expected


@pytest.mark.parametrize(
    'input,expected,expectation',
    [
        ({1, 2, 3}, [1, 2, 3], does_not_raise()),
        ((3.22, 2.11, 1.22), [3.22, 2.11, 1.22], does_not_raise()),
    ]
)
def test_frozenset(input, expected, expectation):

    @dataclass
    class MyClass(JSONWizard):
        num_set: FrozenSet[int]
        any_set: frozenset

    expected = sorted(expected)

    input_set = frozenset(input)
    c = MyClass(num_set=input_set, any_set=input_set)

    with expectation:
        result = c.to_dict()
        log.debug('Parsed object: %r', result)

        assert all(key in result for key in ('num_set', 'any_set'))
        assert isinstance(result['num_set'], (list, tuple))
        assert isinstance(result['any_set'], (list, tuple))

        assert sorted(result['num_set']) == expected
        assert sorted(result['any_set']) == expected


@pytest.mark.parametrize(
    'input,expected,expectation',
    [
        ({1, 2, 3}, [1, 2, 3], does_not_raise()),
        ((3.22, 2.11, 1.22), [3.22, 2.11, 1.22], does_not_raise()),
    ]
)
def test_deque(input, expected, expectation):

    @dataclass
    class MyQClass(JSONWizard):
        num_deque: deque[int]
        any_deque: deque

    input_deque = deque(input)
    c = MyQClass(num_deque=input_deque, any_deque=input_deque)

    with expectation:
        result = c.to_dict()
        log.debug('Parsed object: %r', result)

        assert all(key in result for key in ('num_deque', 'any_deque'))
        assert isinstance(result['num_deque'], list)
        assert isinstance(result['any_deque'], list)

        assert result['num_deque'] == expected
        assert result['any_deque'] == expected


@pytest.mark.parametrize(
    'input,expectation',
    [
        # ideally, an error should be raised
        ('testing', does_not_raise()),
        ('e1', does_not_raise()),
        # ideally, an error should be raised
        (False, does_not_raise()),
        (0, does_not_raise()),
    ]
)
def test_literal(input, expectation):

    @dataclass
    class MyClass(JSONWizard):
        class _(JSONWizard.Meta):
            dump_case = 'PASCAL'

        my_lit: Literal['e1', 'e2', 0]

    c = MyClass(my_lit=input)
    expected = {'MyLit': input}

    with expectation:
        actual = c.to_dict()
        assert actual == expected
        log.debug('Parsed object: %r', actual)


@pytest.mark.parametrize(
    'input,expectation',
    [
        (UUID('12345678-1234-1234-1234-1234567abcde'), does_not_raise()),
        (UUID('{12345678-1234-5678-1234-567812345678}'), does_not_raise()),
        (UUID('12345678123456781234567812345678'), does_not_raise()),
        (UUID('urn:uuid:12345678-1234-5678-1234-567812345678'), does_not_raise()),
    ]
)
def test_uuid(input, expectation):

    @dataclass
    class MyClass(JSONWizard):
        class _(JSONWizard.Meta):
            dump_case = 'Snake'

        my_id: UUID

    c = MyClass(my_id=input)
    expected = {'my_id': input.hex}

    with expectation:
        actual = c.to_dict()
        assert actual == expected
        log.debug('Parsed object: %r', actual)


@pytest.mark.parametrize(
    'input,expectation',
    [
        (timedelta(seconds=12345), does_not_raise()),
        (timedelta(hours=1, minutes=32), does_not_raise()),
        (timedelta(days=1, minutes=51, seconds=7), does_not_raise()),
    ]
)
def test_timedelta(input, expectation):

    @dataclass
    class MyClass(JSONWizard):
        class _(JSONWizard.Meta):
            dump_case = 'Snake'

        my_td: timedelta

    c = MyClass(my_td=input)
    expected = {'my_td': str(input)}

    with expectation:
        actual = c.to_dict()
        assert actual == expected
        log.debug('Parsed object: %r', actual)


@pytest.mark.parametrize(
    'input,expectation',
    [
        ({}, pytest.raises(ParseError)),
        ({'key': 'value'}, pytest.raises(ParseError)),
        ({'my_str': 'test', 'my_int': 2,
          'my_bool': True, 'other_key': 'testing'}, does_not_raise()),
        ({'my_str': 3}, pytest.raises(ParseError)),
        ({'my_str': 'test', 'my_int': 'test', 'my_bool': True}, does_not_raise()),
        ({'my_str': 'test', 'my_int': 2, 'my_bool': True}, does_not_raise()),
    ]
)
def test_typed_dict(input, expectation):

    class MyDict(TypedDict):
        my_str: str
        my_bool: bool
        my_int: int

    @dataclass
    class MyClass(JSONWizard):
        my_typed_dict: MyDict

    c = MyClass(my_typed_dict=input)

    with expectation:
        result = c.to_dict()
        log.debug('Parsed object: %r', result)


def test_using_dataclass_in_dict():
    """
    Using dataclass in a dictionary (i.e., dict[str, Test])
    works as expected.

    See https://github.com/rnag/dataclass-wizard/issues/159
    """
    @dataclass
    class Test:
        field: str

    @dataclass
    class Config:
        tests: dict[str, Test]

    config = {"tests": {"test_a": {"field": "a"}, "test_b": {"field": "b"}}}

    assert fromdict(Config, config) == Config(
        tests={'test_a': Test(field='a'),
               'test_b': Test(field='b')})


def test_bytes_and_bytes_array_are_supported():
    """Confirm dump with `bytes` and `bytesarray` is supported."""

    @dataclass
    class Foo(JSONWizard):
        b: bytes = None
        barray: bytearray = None
        s: str = None

    data = {'b': 'AAAA', 'barray': 'SGVsbG8sIFdvcmxkIQ==', 's': 'foobar'}

    foo = Foo(b=b64decode('AAAA'),
              barray=bytearray(b'Hello, World!'),
              s='foobar')

    assert foo.to_dict() == data



import json
from dataclasses import Field
from typing import AnyStr, dataclass_transform, Collection, Sequence

from ..abstractions import AbstractEnvWizard, E
from ..bases_meta import BaseEnvWizardMeta
from ..enums import LetterCase
from ..errors import ParseError
from ..type_def import JSONObject, Encoder, EnvFileType


@dataclass_transform(kw_only_default=True)
class EnvWizard(AbstractEnvWizard):
    __slots__ = ()

    class Meta(BaseEnvWizardMeta):

        __slots__ = ()

        # Class attribute to enable detection of the class type.
        __is_inner_meta__ = True

        def __init_subclass__(cls):
            # Set the `__init_subclass__` method here, so we can ensure it
            # doesn't run for the `EnvWizard.Meta` class.
            return cls._init_subclass()

    __fields__: dict[str, Field]

    def to_dict(self: E,
                *,
                dict_factory=dict,
                exclude: Collection[str] | None = None,
                skip_defaults: bool | None = None,
                ) -> JSONObject: ...

    def to_json(self: E, *,
                encoder: Encoder = json.dumps,
                **encoder_kwargs) -> AnyStr: ...

    # stub for type hinting purposes.
    def __init__(self, *,
                 _env_file: EnvFileType = None,
                 _reload: bool = False,
                 _env_prefix:str=None,
                 _secrets_dir:EnvFileType | Sequence[EnvFileType]=None,
                 **init_kwargs) -> None: ...

    def __init_subclass__(cls, *, reload_env: bool = False,
                          debug: bool = False,
                          key_transform=LetterCase.NONE): ...

    @classmethod
    def _create_methods(cls) -> None: ...


def _add_missing_var(missing_vars: list,
                     name: str,
                     env_prefix: str | None,
                     var_name: str | None,
                     tp: type) -> None: ...


def _handle_parse_error(e: ParseError,
                        cls: type,
                        name: str,
                        env_prefix: str | None,
                        var_name: str | None): ...

def _get_var_name(name: str,
                  env_prefix: str | None,
                  var_name: str | None) -> str: ...



import base64
import json
from collections import deque
from dataclasses import field, is_dataclass
from datetime import datetime, date, timezone
from typing import Optional, Union, NamedTuple, Literal

import pytest

from dataclass_wizard import asdict, fromdict, Alias, DataclassWizard
from dataclass_wizard.models import CatchAll
from dataclass_wizard.errors import ParseError, MissingFields
from .models import TN, CN, ContTF, ContTT, ContAllReq, Sub2, TNReq
from .utils_env import assert_unordered_equal
from tests._typing import *


def test_nested_union_with_complex_types_in_containers():

    class Sub(DataclassWizard):
        test: str

    class MyClass(DataclassWizard):
        class _(DataclassWizard.Meta):
            case = 'CAMEL'
            auto_assign_tags = True

        # noinspection PyDataclass
        my_bool: dict[str, tuple[Optional[bool], ...]] = Alias('Boolean-Dict')
        nestedUnionWithClass: list[Union[str, Sub]]

    d = {'Boolean-Dict': {'test': [None, 'tRuE', None]}, 'nestedUnionWithClass': [123, {'__tag__': 'Sub', 'test': 'value'}]}

    c = fromdict(MyClass, d)
    assert c == MyClass(
        my_bool={'test': (None, True, None)},
        nestedUnionWithClass=['123', Sub(test='value')],
    )

    new_dict = asdict(c)
    assert new_dict == {
        'Boolean-Dict': {'test': [None, True, None]},
        'nestedUnionWithClass': ['123', {'test': 'value', '__tag__': 'Sub'}]
    }


def test_named_tuples_with_optionals_in_container():

    class NTAllOptionals(NamedTuple):
        a: str = 'test'
        b: int = 1
        c: bool = True

    class NTOneOptional(NamedTuple):
        my_int: int
        my_bool: bool = False

    class MyClass(DataclassWizard):
        class _(DataclassWizard.Meta):
            case = 'PASCAL'

        nt_all_opts: dict[str, set[NTAllOptionals]]
        nt_one_opt: list[NTOneOptional]

    d = {'NtOneOpt': [['123']], 'NtAllOpts': {'k': [[]]}}

    with pytest.raises(ParseError) as e:
        _ = fromdict(MyClass, {'NtOneOpt': [{}], 'NtAllOpts': {'k': [[]]}})

    # TODO
    # assert '`dict` input is not supported for NamedTuple, use a dataclass instead' in str(e.value)

    c = fromdict(MyClass, d)
    assert c == MyClass(nt_all_opts={'k': {NTAllOptionals()}},
                        nt_one_opt=[NTOneOptional(my_int=123, my_bool=False)])

    new_dict = asdict(c)
    assert new_dict == {
        'NtAllOpts': {
            'k': [['test', 1, True]]
        },
        'NtOneOpt': [
            [123, False]
        ]
    }


def test_typed_dict_with_optionals_and_read_only_in_container():

    class TDWithReadOnlyAndOpts(TypedDict):
        a: int
        b: NotRequired[str]
        c: NotRequired[bool]
        d: ReadOnly[float]

    class MyClass(DataclassWizard):
        my_td: dict[str, tuple[bool, deque[TDWithReadOnlyAndOpts]]]

    td_value = []
    d = {'my_td': {'test': ['True', [td_value]]}}

    with pytest.raises(ParseError) as e:
        _ = fromdict(MyClass, d)
    assert 'Expected a type {}, got list' in str(e.value)

    td_value = {'d': '23'}
    d = {'my_td': {'test': ['True', [td_value]]}}

    with pytest.raises(ParseError) as e:
        _ = fromdict(MyClass, d)
    assert 'Missing required key: \'a\'' in str(e.value)

    td_value = {'d': '23', 'a': '1', 'c': '0'}
    d = {'my_td': {'test': ['True', [td_value]]}}

    c = fromdict(MyClass, d)
    assert c == MyClass(my_td={'test': (True, deque([{'a': 1, 'c': False, 'd': 23.0}]))})

    new_dict = asdict(c)
    assert new_dict == {'my_td': {'test': [True, [{'a': 1, 'c': False, 'd': 23.0}]]}}


def test_literal_in_container():
    class MyClass(DataclassWizard):
        my_literal_dict: dict[str, tuple[int, frozenset[Literal['Aa', 'Bb']]]]

    d = {'my_literal_dict': {'test': ['123', ['aa', 'Bb']]}}

    with pytest.raises(ParseError) as e:
        _ = MyClass.from_dict(d)
    assert 'Value not in expected Literal values' in str(e.value)

    d = {'my_literal_dict': {'test': ['123', ['Bb', 'Aa']]}}
    c = MyClass.from_dict(d)
    # noinspection PyTypeChecker
    assert c == MyClass(my_literal_dict={'test': (123, frozenset({'Aa', 'Bb'}))})

    new_dict = c.to_dict()
    assert_unordered_equal(new_dict, {'my_literal_dict': {'test': [123, ['Aa', 'Bb']]}})


def test_decode_date_and_datetime_from_numeric_and_string_timestamp_and_iso_format():

    class MyClass(DataclassWizard):
        my_value: float
        my_dt: datetime
        another_date: date

    d = {"another_date": "2021-12-17", "my_dt": "2022-04-27T16:30:45Z", "my_value": 1.23}
    expected_obj = MyClass(my_value=1.23,
                           my_dt=datetime(2022, 4, 27, 16, 30, 45, tzinfo=timezone.utc),
                           another_date=date(2021, 12, 17))
    expected_dict = {'my_value': 1.23, 'my_dt': '2022-04-27T16:30:45Z', 'another_date': '2021-12-17'}

    c1 = MyClass.from_dict(d)

    d = {"another_date": "1639763585", "my_dt": "1651077045", "my_value": '1.23'}
    c2 = MyClass.from_dict(d)

    d = {"another_date": 1639763585, "my_dt": 1651077045, "my_value": '1.23'}
    c3 = MyClass.from_dict(d)

    assert c1 == c2 == c3 == expected_obj
    assert c1.to_dict() == c2.to_dict() == c3.to_dict() == expected_dict


def test_json_bytes_roundtrip():
    class C(DataclassWizard):
        b: bytes

    raw = b'testing 123'
    c = C(b=raw)

    s = c.to_json()
    d = json.loads(s)

    # JSON representation should be base64 ascii string
    assert isinstance(d['b'], str)
    assert base64.b64decode(d['b']) == raw

    c2 = C.from_json(s)
    assert c2.b == raw


def test_json_bytearray_roundtrip():
    class C(DataclassWizard):
        b: bytearray

    raw = bytearray(b'abc')
    s = C(b=raw).to_json()
    c2 = C.from_json(s)
    assert c2.b == raw


def test_from_dict_bytes_requires_base64_str():
    class C(DataclassWizard):
        b: bytes

    with pytest.raises(ParseError):
        C.from_dict({'b': 'raw'})


def test_from_dict_bytes_accepts_bytes_and_bytearray():
    class C(DataclassWizard):
        b: bytes

    assert C.from_dict({'b': b'raw'}).b == b'raw'
    assert C.from_dict({'b': bytearray(b'raw')}).b == b'raw'


@pytest.mark.skipif(not PY310_OR_ABOVE, reason='requires Python 3.10 or higher')
def test_kw_only_fields_and_kw_only_catchall():
    from dataclasses import KW_ONLY

    class Sub(DataclassWizard):
        test: str
        _: KW_ONLY
        my_int: int

    class Parent(DataclassWizard, kw_only=True):
        opt: Optional[Sub]
        extras: CatchAll
        my_bool: bool = field(kw_only=False)

    p = Parent.from_dict({'my_bool': 'false',
                          'opt': {'my_int': '321', 'test': 123},
                          'hello': 0, 'world': 1})
    assert p == Parent(opt=Sub(test='123', my_int=321),
                       extras={'hello': 0, 'world': 1},
                       my_bool=False)

    assert p.to_dict() == {'my_bool': False,
                           'opt': {'test': '123', 'my_int': 321},
                           'hello': 0, 'world': 1}


def test_lazy_codegen_does_not_poison_subclasses():
    class A(DataclassWizard):
        a: int = 1

    class B(A):
        b: int = 2

    class C(B):
        c: int = 3

    assert is_dataclass(A) and is_dataclass(B) and is_dataclass(C)

    assert getattr(A.from_dict, '__func__', None) is fromdict
    assert getattr(B.from_dict, '__func__', None) is fromdict
    assert getattr(C.from_dict, '__func__', None) is fromdict

    a = A.from_dict({'a': '10'})
    assert type(a) is A
    assert a.a == 10

    b = B.from_dict({'a': '10', 'b': '20'})
    assert type(b) is B
    assert (b.a, b.b) == (10, 20)

    c = C.from_dict({'a': '10', 'b': '20', 'c': '30'})
    assert type(c) is C
    assert (c.a, c.b, c.c) == (10, 20, 30)

    # Ensure each class dumps its own fields correctly
    assert a.to_dict() == {'a': 10}
    assert b.to_dict() == {'a': 10, 'b': 20}
    assert c.to_dict() == {'a': 10, 'b': 20, 'c': 30}

    assert A.from_dict is not B.from_dict
    assert B.from_dict is not C.from_dict
    assert A.from_dict is not C.from_dict
    assert A.from_dict is not fromdict
    assert B.from_dict is not fromdict
    assert C.from_dict is not fromdict


class ContDict(DataclassWizard):
    class _(DataclassWizard.Meta):
        namedtuple_as_dict = True

    tn: TN
    cn: CN


class ContDictReq(DataclassWizard):
    class _(DataclassWizard.Meta):
        namedtuple_as_dict = True

    tn: TNReq


def test_namedtuple_dict_mode_roundtrip_and_defaults():
    o = ContDict.from_dict({"tn": {"a": 1}, "cn": {"a": 3}})
    assert o.tn == TN(a=1, b=2)
    assert o.cn == CN(a=3, b=2)

    d = o.to_dict()
    assert d == {"tn": {"a": 1, "b": 2}, "cn": {"a": 3, "b": 2}}


def test_namedtuple_no_field_defaults_dict_mode_roundtrip():
    o = ContDictReq.from_dict({"tn": {"b": 1, "a": 3}})

    assert o.tn == TN(a=3, b=1)

    d = o.to_dict()
    assert d == {'tn': {'a': 3, 'b': 1}}


def test_namedtuple_no_field_defaults_dict_mode_missing_required_raises():
    with pytest.raises(MissingFields, match=r'`TNReq.__init__\(\)` missing required fields') as e:
        _ = ContDictReq.from_dict({"tn": {"a": 1}})

    assert e.value.missing_fields == ['b']


def test_namedtuple_dict_mode_missing_required_raises():
    with pytest.raises(MissingFields, match=r'`TN\.__init__\(\)` missing required fields') as e:
        ContDict.from_dict({"tn": {"b": 9}, "cn": {"a": 1}})

    assert e.value.missing_fields == ['a']


class ContList(DataclassWizard):
    class _(DataclassWizard.Meta):
        namedtuple_as_dict = False

    tn: TN
    cn: CN


def test_namedtuple_list_mode_roundtrip_and_defaults():
    o = ContList.from_dict({"tn": [1], "cn": [3]})
    assert o.tn == TN(a=1, b=2)
    assert o.cn == CN(a=3, b=2)

    d = o.to_dict()
    assert d == {"tn": [1, 2], "cn": [3, 2]}


def test_namedtuple_list_mode_rejects_dict_input_with_clear_error():
    with pytest.raises(ParseError, match=r"Dict input is not supported for NamedTuple fields in list mode.*list.*Meta\.namedtuple_as_dict = True"):
        ContList.from_dict({"tn": {"a": 1}, "cn": {"a": 3}})


def test_namedtuple_dict_mode_rejects_dict_input_with_clear_error():
    with pytest.raises(ParseError, match=r"List/tuple input is not supported for NamedTuple fields in dict mode.*dict.*Meta\.namedtuple_as_dict = False"):
        ContDict.from_dict({"tn": ['test'], "cn": {"a": 3}})


def test_typeddict_total_false_e2e_dict_roundtrip():
    o = ContTF.from_dict({"td": {"a": 1, "ro": 9}})
    assert o.td == {"a": 1, "ro": 9}

    d = o.to_dict()
    assert d == {"td": {"a": 1, "ro": 9}}


def test_typeddict_total_false_missing_required_raises():
    with pytest.raises(Exception):  # swap to MissingFields/TypeError etc
        ContTF.from_dict({"td": {"b": 2}})


def test_typeddict_total_true_e2e_optional_and_required_keys():
    o = ContTT.from_dict({"td": {"a": 1, "ro": 9}})
    assert o.td == {"a": 1, "ro": 9}

    d = o.to_dict()
    assert d == {"td": {"a": 1, "ro": 9}}

    with pytest.raises(Exception):
        ContTT.from_dict({"td": {"a": 1}})   # missing ro


def test_typeddict_all_required_e2e_inline_path():
    o = ContAllReq.from_dict({"td": {"x": 1, "y": "ok"}})
    assert o.td == {"x": 1, "y": "ok"}

    d = o.to_dict()
    assert d == {"td": {"x": 1, "y": "ok"}}

    with pytest.raises(Exception):
        ContAllReq.from_dict({"td": {"x": 1}})  # missing y


def test_union_codegen_cache_nested_union_roundtrip_and_dump_error():
    class MyClass(DataclassWizard):
        class _(DataclassWizard.Meta):
            unsafe_parse_dataclass_in_union = True

        complex_tp: 'list[int | Sub2] | list[int | str]'

    # First: pick the arm list[int|Sub2]
    o1 = MyClass.from_dict({"complex_tp": [{"my_float": "123."}, 7]})
    assert o1 == MyClass(complex_tp=[Sub2(my_float=123.0), 7])

    d1 = o1.to_dict()
    assert d1 == {"complex_tp": [{"my_float": 123.0}, 7]}

    # Second: pick the other arm list[int|str]
    # If inner-union codegen caching is wrong, this is where it tends to misbehave.
    o2 = MyClass.from_dict({"complex_tp": ["hello", 9]})
    assert o2 == MyClass(complex_tp=["hello", 9])

    d2 = o2.to_dict()
    assert d2 == {"complex_tp": ["hello", 9]}

    o1.complex_tp = '123'
    with pytest.raises(ParseError, match=r"Failed to dump field `complex_tp` in class `.*MyClass`"):
        _ = o1.to_dict()



def to_camel_case(string: str) -> str: ...
def to_pascal_case(string): ...
def to_lisp_case(string: str) -> str: ...
def to_snake_case(string: str) -> str: ...
def replace_multi_with_single(string: str, char: str = ...) -> str: ...



__all__ = ['Buffer', 'Unpack', 'PyForwardRef', 'PyProtocol', 'PyDeque', 'PyTypedDict', 'PyRequired', 'PyNotRequired', 'PyReadOnly', 'PyLiteralString', 'FrozenKeys', 'DefFactory', 'NoneType', 'ExplicitNullType', 'ExplicitNull', 'JSONList', 'JSONObject', 'ListOfJSONObject', 'JSONValue', 'FileType', 'EnvFileType', 'StrCollection', 'ParseFloat', 'Encoder', 'FileEncoder', 'Decoder', 'FileDecoder', 'NUMBERS', 'T', 'E', 'U', 'M', 'NT', 'DT', 'DD', 'N', 'S', 'LT', 'LSQ', 'FREF', 'dataclass_transform', 'UNSET', 'META', 'ENV_META', '_META', '_ENV_META']

import typing
from collections.abc import Buffer as Buffer
from datetime import date, datetime, time
from enum import Enum
from os import PathLike
from typing import ClassVar
from typing import Deque as PyDeque
from typing import ForwardRef as PyForwardRef
from typing import LiteralString as PyLiteralString
from typing import NotRequired as PyNotRequired
from typing import Protocol as PyProtocol
from typing import ReadOnly as PyReadOnly
from typing import Required as PyRequired
from typing import TypedDict as PyTypedDict
from typing import Unpack as Unpack
from typing import dataclass_transform as dataclass_transform

from _typeshed import SupportsRead, SupportsWrite

from ._bases import AbstractEnvMeta, AbstractMeta

FrozenKeys = frozenset[str]
JSONList = list[typing.Any]
JSONObject = dict[str, typing.Any]
ListOfJSONObject = list[JSONObject]
NoneType = type(None)

FileType = typing.Union[str, bytes, PathLike, int]
EnvFileType = typing.Union[bool, FileType, typing.Iterable[FileType], None]
JSONValue = typing.Union[None, str, bool, int, float, JSONList, JSONObject]
ParseFloat = typing.Callable[[str], typing.Any]
N = typing.Union[int, float]
StrCollection = typing.Union[str, typing.Collection[str]]

# Create a generic variable that can be 'AbstractMeta', or any subclass.
# Full word as `M` is already defined in another module
_META = typing.TypeVar('_META', bound=AbstractMeta)
# Use `type` here explicitly, because we will never have an `META_` object.
META = type[_META]

# Create a generic variable that can be 'AbstractMeta', or any subclass.
# Full word as `M` is already defined in another module
_ENV_META = typing.TypeVar('_ENV_META', bound=AbstractEnvMeta)
# Use `type` here explicitly, because we will never have an `META_` object.
ENV_META = type[_ENV_META]

NUMBERS: tuple
T = typing.TypeVar('T')
T_co = typing.TypeVar('T_co', covariant=True)

@typing.type_check_only
class DefFactory(typing.Protocol[T_co]):
    def __call__(self) -> T_co: ...

E = typing.TypeVar('E', bound=Enum)
U: typing.TypeVar
M: typing.TypeVar
NT: typing.TypeVar
DT = typing.TypeVar('DT', date, time, datetime)
DD: typing.TypeVar
S: typing.TypeVar
LT: typing.TypeVar
LSQ: typing.TypeVar
FREF = typing.TypeVar('FREF', str, PyForwardRef)

class _UnsetType: ...
UNSET: _UnsetType

class ExplicitNullType:
    _instance: ClassVar[ExplicitNullType] = ...
    @classmethod
    def __init__(cls) -> None: ...
    def __bool__(self) -> bool: ...
ExplicitNull: ExplicitNullType

class Encoder(typing.Protocol):
    def __call__(self, obj: JSONObject | JSONList, /, *args: typing.Any, **kwargs: typing.Any) -> str: ...
    @classmethod
    def __subclasshook__(cls, other): ...
    def __init__(self, *args, **kwargs) -> None: ...

class FileEncoder(typing.Protocol):
    def __call__(self, obj: JSONObject | JSONList,
                 file: SupportsWrite[str],
                 /, *args: typing.Any, **kwargs: typing.Any) -> None: ...
    @classmethod
    def __subclasshook__(cls, other): ...
    def __init__(self, *args, **kwargs) -> None: ...

class Decoder(typing.Protocol):
    def __call__(self, s: str | bytes | bytearray, /, *args: typing.Any, **kwargs: typing.Any) -> JSONObject | ListOfJSONObject: ...
    @classmethod
    def __subclasshook__(cls, other): ...
    def __init__(self, *args, **kwargs) -> None: ...

class FileDecoder(typing.Protocol):
    def __call__(self, file: SupportsRead[str | bytes], /, *args: typing.Any, **kwargs: typing.Any) -> JSONObject | ListOfJSONObject: ...
    @classmethod
    def __subclasshook__(cls, other): ...
    def __init__(self, *args, **kwargs) -> None: ...

# Names in __all__ with no definition:
#   Buffer
#   DefFactory
#   EnvFileType
#   FileType
#   FrozenKeys
#   JSONList
#   JSONObject
#   JSONValue
#   ListOfJSONObject
#   N
#   NoneType
#   ParseFloat
#   PyDeque
#   PyForwardRef
#   PyLiteralString
#   PyNotRequired
#   PyProtocol
#   PyReadOnly
#   PyRequired
#   PyTypedDict
#   StrCollection
#   Unpack
#   dataclass_transform



"""
Common test fixtures and utilities.
"""
from dataclasses import dataclass
from uuid import UUID

import pytest


# Ref: https://docs.pytest.org/en/6.2.x/example/parametrize.html#parametrizing-conditional-raising
from contextlib import nullcontext as does_not_raise


@dataclass
class SampleClass:
    """Sample dataclass model for various test scenarios."""
    f1: str
    f2: int


class MyUUIDSubclass(UUID):
    """
    Simple UUID subclass that calls :meth:`hex` when ``str()`` is invoked.
    """

    def __str__(self):
        return self.hex


@pytest.fixture
def mock_log(caplog):
    caplog.set_level('INFO', logger='dataclass_wizard')
    return caplog

@pytest.fixture
def mock_debug_log(caplog):
    caplog.set_level('DEBUG', logger='dataclass_wizard')
    return caplog



from collections import deque
from dataclasses import field
from datetime import datetime, date, timezone
from typing import Optional, Union, NamedTuple, Literal

import pytest

from dataclass_wizard import (Alias, DataclassWizard,
                              EnvWizard, AliasPath)
from dataclass_wizard.env import env_config
from dataclass_wizard.models import CatchAll
from dataclass_wizard.errors import ParseError, MissingVars, MissingFields
from ..models import TN, CN, EnvContTF, EnvContTT, EnvContAllReq, Sub2

from ..utils_env import envsafe, from_env, assert_unordered_equal
from tests._typing import *


def test_none_is_deserialized():
    class MyClass(EnvWizard):
        test: Union[int, None]

    _input = {'TEST': 123}
    env = envsafe(_input)
    assert env == {'TEST': '123'}
    c = from_env(MyClass, _input)
    assert c.raw_dict() == {'test': 123}

    _input = {'TEST': None}  # None -> 'null'
    env = envsafe(_input)
    assert env == {'TEST': 'null'}
    c = from_env(MyClass, _input)
    assert c.raw_dict() == {'test': None}


def test_nested_union_with_complex_types_in_containers():
    class Sub(DataclassWizard):
        test: str

    class MyClass(EnvWizard):
        class _(EnvWizard.Meta):
            case = 'CAMEL'
            unsafe_parse_dataclass_in_union = True

        my_bool: dict[str, tuple[Optional[bool], ...]] = Alias(env='Boolean-Dict')
        unionInListWithClass: list[Union[str, Sub, None]]
        unionWithClass: Union[str, Sub]
        opt_class: Optional[Sub]


    _input = {'Boolean-Dict': {'v': [None, 'tRuE', None]},
              'UNION_IN_LIST_WITH_CLASS': ['s', {'test': 'hello'}, 'value'],
              'UNION_WITH_CLASS': {'test': 'world'},
              'OPT_CLASS': {'test': 'value'}}

    c = from_env(MyClass, _input)
    assert c.raw_dict() == {'my_bool': {'v': (None, True, None)},
                            'unionInListWithClass': ['s', Sub(test='hello'), 'value'],
                            'unionWithClass': Sub(test='world'), 'opt_class': Sub(test='value')}

    _input = {'Boolean-Dict': {'v': []},
              'UNION_IN_LIST_WITH_CLASS': ['', {'test': '123'}, None],
              'UNION_WITH_CLASS': 's',
              'OPT_CLASS': None}

    c = from_env(MyClass, _input)
    assert c.raw_dict() == {'my_bool': {'v': ()},
                            'unionInListWithClass': ['', Sub(test='123'), None], 'unionWithClass': 's',
                            'opt_class': None}


def test_named_tuples_with_optionals_in_container():

    class NTAllOptionals(NamedTuple):
        a: str = 'test'
        b: int = 1
        c: bool = True

    class NTOneOptional(NamedTuple):
        my_int: int
        my_bool: bool = False

    class MyClass(EnvWizard):
        class _(EnvWizard.Meta):
            load_case = 'FIELD_FIRST'

        nt_all_opts: dict[str, set[NTAllOptionals]]
        nt_one_opt: list[NTOneOptional]

    _input = {'nt_one_opt': [['123']], 'nt_all_opts': {'k': [[]]}}

    with pytest.raises(ParseError) as e:
        _ = from_env(MyClass, {'nt_one_opt': [{}], 'nt_all_opts': {'k': [[]]}})

    # TODO
    # assert '`dict` input is not supported for NamedTuple, use a dataclass instead' in str(e.value)

    c = from_env(MyClass, _input)

    assert c.raw_dict() == {
        'nt_all_opts': {
            'k': {NTAllOptionals()},
        },
        'nt_one_opt': [
            NTOneOptional(my_int=123, my_bool=False)
        ],
    }


def test_typed_dict_with_optionals_and_read_only_in_container():

    class TDWithReadOnlyAndOpts(TypedDict):
        a: int
        b: NotRequired[str]
        c: NotRequired[bool]
        d: ReadOnly[float]

    class MyClass(EnvWizard):
        my_td: dict[str, tuple[bool, deque[TDWithReadOnlyAndOpts]]]

    td_value = []
    d = {'my_td': {'test': ['True', [td_value]]}}

    with pytest.raises(ParseError) as e:
        _ = from_env(MyClass, d)
    assert 'Expected a type {}, got list' in str(e.value)

    td_value = {'d': '23'}
    d = {'my_td': {'test': ['True', [td_value]]}}

    with pytest.raises(ParseError) as e:
        _ = from_env(MyClass, d)
    assert 'Missing required key: \'a\'' in str(e.value)

    td_value = {'d': '23', 'a': '1', 'c': '0'}
    d = {'my_td': {'test': ['True', [td_value]]}}

    c = from_env(MyClass, d)
    assert c == MyClass(my_td={'test': (True, deque([{'a': 1, 'c': False, 'd': 23.0}]))})

    new_dict = c.raw_dict()
    assert new_dict == {'my_td': {'test': (True, deque([{'a': 1, 'c': False, 'd': 23.0}]))}}


def test_literal_in_container():
    class MyClass(EnvWizard):
        my_literal_dict: dict[str, tuple[int, frozenset[Literal['Aa', 'Bb']]]]

    d = {'my_literal_dict': {'test': ['123', ['aa', 'Bb']]}}

    with pytest.raises(ParseError) as e:
        _ = from_env(MyClass, d)
    assert 'Value not in expected Literal values' in str(e.value)

    d = {'my_literal_dict': {'test': ['123', ['Bb', 'Aa']]}}
    c = from_env(MyClass, d)
    # noinspection PyTypeChecker
    assert c == MyClass(my_literal_dict={'test': (123, frozenset({'Aa', 'Bb'}))})

    new_dict = c.to_dict()
    assert_unordered_equal(new_dict, {'my_literal_dict': {'test': [123, ['Aa', 'Bb']]}})


def test_decode_date_and_datetime_from_numeric_and_string_timestamp_and_iso_format():

    class MyClass(EnvWizard):
        my_value: float
        my_dt: datetime
        another_date: date

    d = {"another_date": "2021-12-17", "my_dt": "2022-04-27T16:30:45Z", "my_value": 1.23}

    expected_obj = MyClass(my_value=1.23,
                           my_dt=datetime(2022, 4, 27, 16, 30, 45, tzinfo=timezone.utc),
                           another_date=date(2021, 12, 17))
    expected_dict = {'my_value': 1.23, 'my_dt': '2022-04-27T16:30:45Z', 'another_date': '2021-12-17'}

    c1 = from_env(MyClass, d)

    d = {"another_date": "1639763585", "my_dt": "1651077045", "my_value": '1.23'}
    c2 = from_env(MyClass, d)

    d = {"another_date": 1639763585, "my_dt": 1651077045, "my_value": '1.23'}
    c3 = from_env(MyClass, d)

    assert c1 == c2 == c3 == expected_obj
    assert c1.to_dict() == c2.to_dict() == c3.to_dict() == expected_dict


def test_field_to_env_load():
    """Meta field `field_to_env_load` usage."""

    class MyClass(EnvWizard):
        class _(EnvWizard.Meta):
            field_to_env_load = {'my_value': 'MyVal', 'other_key': ('INT1', 'INT2')}

        my_value: float
        other_key: int = 3

    env = {'MyVal': '1.23', 'INT2': '7.0'}

    c = from_env(MyClass, env)

    assert c == MyClass(my_value=1.23, other_key=7)
    assert c.raw_dict() == {'my_value': 1.23, 'other_key': 7}


def test_env_bytes_from_string_is_utf8():
    class C(EnvWizard):
        b: bytes

    c = from_env(C, {'B': 'testing 123'})
    assert c.b == b'testing 123'


def test_env_bytearray_from_string_is_utf8():
    class C(EnvWizard):
        b: bytearray

    c = from_env(C, {'B': 'testing 123'})
    assert c.b == bytearray(b'testing 123')


def test_env_bytearray_from_bytes_and_list():
    class C(EnvWizard):
        b: bytearray

    assert C(__env__=env_config(mapping={'B': b'abc'})).b == bytearray(b'abc')
    assert C(__env__=env_config(mapping={'B': [1, 2, 3]})).b == bytearray([1, 2, 3])


def test_env_with_no_init_fields():
    """Test EnvWizard` subclass with no *init-only *dataclass fields."""
    class E1(EnvWizard):
        ...

    assert E1().raw_dict() == {}

    class E2(EnvWizard):
        my_field: str = field(init=False)

        def __post_init__(self, __env__=None):
            self.my_field = '123'

    assert E2().raw_dict() == {'my_field': '123'}


def test_env_with_catch_all():
    class E(EnvWizard):
        field: CatchAll = None

    class MyDict(dict):
        def __bool__(self):
            return True

    e = E(__env__={'mapping': MyDict()})
    assert e.field is None
    assert e.to_dict() == {}

    e = from_env(E, {'k1': 'v1', 'k2': ['v2']})
    assert e.to_dict() == e.field == {'k1': 'v1', 'k2': '["v2"]'}

    class E(EnvWizard):
        field: CatchAll

    e = from_env(E, {'k1': 'v1', 'k2': ['v2']})
    assert e.to_dict() == e.field == {'k1': 'v1', 'k2': '["v2"]'}

    class E(EnvWizard):
        my_int: int
        extras: CatchAll

    e = from_env(E, {'MY_INT': '123',
                     'k1': 'v1',
                     'k2': ['v2']})

    assert e.extras == {'k1': 'v1', 'k2': '["v2"]'}
    assert e.raw_dict() == {'my_int': 123, 'extras': {'k1': 'v1', 'k2': '["v2"]'}}
    assert e.to_dict() == {'my_int': 123, 'k1': 'v1', 'k2': '["v2"]'}


def test_env_precedence_env_only():
    class E(EnvWizard):
        my_value: float

        class _(EnvWizard.Meta):
            env_precedence = 'ENV_ONLY'
            # contains `MY_VALUE=1.23`
            env_file = '.env.test'

    with pytest.raises(MissingVars) as e:
        _ = E()

    assert e.value.fields.lstrip() == '- my_value -> MY_VALUE'

    e = from_env(E, {'MY_VALUE': '3.21'})
    assert e.raw_dict() == {'my_value': 3.21}


def test_env_load_case_strict():
    class E(EnvWizard):
        my_value: float

        class _(EnvWizard.Meta):
            load_case = 'STRICT'

    with pytest.raises(MissingVars) as e:
        _ = from_env(E, {'my_value': 3.21})

    assert e.value.fields.lstrip() == '- my_value -> my_value'

    e = E(my_value='3.21')
    assert e.raw_dict() == {'my_value': 3.21}


def test_env_alias_path_required():
    class E(EnvWizard):
        my_value: float = AliasPath('a.b.c')

    e = E(my_value=3.21)
    assert e.raw_dict() == {'my_value': 3.21}

    e = E(__env__={'mapping': {'a': {'b': {'c': '2.22'}}}})
    assert e.raw_dict() == {'my_value': 2.22}

    e = from_env(E, {'a': {'b': {'c': '1.11'}}})
    assert e.raw_dict() == {'my_value': 1.11}

    with pytest.raises(ParseError) as e:
        _ = from_env(E, {'a': {'b': {'z': '1.11'}}})
    assert e.value.kwargs['current_path'] == "'c'"

    with pytest.raises(ParseError) as e:
        _ = from_env(E, {'a': []})
    assert str(e.value.base_error) == 'Invalid path'


def test_env_alias_path_with_default_value():
    class E(EnvWizard):
        my_value: list[float] = AliasPath('a.b.c', default_factory=list)
        another_value: Optional[str] = AliasPath('x.y.z', default=None)

    e = E(my_value=[3.21])
    assert e.raw_dict() == {'my_value': [3.21], 'another_value': None}

    e = E(__env__={'mapping': {'a': {'b': {'c': ['2.22']}},
                                'x': {'y': {'z': 3.333}}}})
    assert e.raw_dict() == {'my_value': [2.22], 'another_value': '3.333'}

    e = from_env(E, {'a': {'b': {'c': ['1.11', '2.', 3.7]}},
                      'x': {'y': {'z': 5.55}}})
    assert e.raw_dict() == {'my_value': [1.11, 2.0, 3.7], 'another_value': '5.55'}

    e = from_env(E, {'a': {'b': {'z': '1.11'}}})
    assert e == E(my_value=[], another_value=None)

    with pytest.raises(ParseError) as e:
        _ = from_env(E, {'a': []})
    assert str(e.value.base_error) == 'Invalid path'


def test_env_alias_path_with_multiple_paths():
    class E2(EnvWizard):
        my_value: float = AliasPath('a.b.c', 'x.y.z[0]')

    e = E2(my_value='3.21')
    assert e.raw_dict() == {'my_value': 3.21}

    e = E2(__env__={'mapping': {'a': {'b': {'c': '2.22'}}}})
    assert e.raw_dict() == {'my_value': 2.22}

    e = from_env(E2, {'a': {'b': {'c': '1.11'}}})
    assert e.raw_dict() == {'my_value': 1.11}

    e = from_env(E2, {'x': {'y': {'z': ['3.33', '4.44']}}})
    assert e == E2(my_value=3.33)

    with pytest.raises(ParseError) as e:
        _ = from_env(E2, {'a': []})
    assert str(e.value.base_error) == 'Invalid path'


def test_namedtuple_dict_mode_roundtrip_and_defaults():
    class EnvContDict(EnvWizard):
        class _(EnvWizard.Meta):
            namedtuple_as_dict = True

        tn: TN
        cn: CN

    o = from_env(EnvContDict, {"tn": {"a": 1}, "cn": {"a": 3}})
    assert o.tn == TN(a=1, b=2)
    assert o.cn == CN(a=3, b=2)

    d = o.to_dict()
    assert d == {"tn": {"a": 1, "b": 2}, "cn": {"a": 3, "b": 2}}


# TODO
# def test_namedtuple_dict_mode_missing_required_raises():
#     with pytest.raises(MissingFields, match=r'`TN\.__init__\(\)` missing required fields') as e:
#         from_env(EnvContDict, {"tn": {"b": 9}, "cn": {"a": 1}})
#
#     assert e.value.missing_fields == ['a']


def test_namedtuple_list_mode_roundtrip_and_defaults():
    class EnvContList(EnvWizard):
        class _(EnvWizard.Meta):
            namedtuple_as_dict = False

        tn: TN
        cn: CN

    o = from_env(EnvContList, {"tn": [1], "cn": [3]})
    assert o.tn == TN(a=1, b=2)
    assert o.cn == CN(a=3, b=2)

    d = o.to_dict()
    assert d == {"tn": [1, 2], "cn": [3, 2]}


# def test_namedtuple_list_mode_rejects_dict_input_with_clear_error():
#     with pytest.raises(ParseError, match=r"Dict input is not supported for NamedTuple fields in list mode.*list.*Meta\.namedtuple_as_dict = True"):
#         from_env(EnvContList, {"tn": {"a": 1}, "cn": {"a": 3}})


def test_typeddict_total_false_e2e_dict_roundtrip():
    o = from_env(EnvContTF, {"td": {"a": 1, "ro": 9}})
    assert o.td == {"a": 1, "ro": 9}

    d = o.to_dict()
    assert d == {"td": {"a": 1, "ro": 9}}


def test_typeddict_total_false_missing_required_raises():
    with pytest.raises(Exception):  # swap to MissingFields/TypeError etc
        from_env(EnvContTF, {"td": {"b": 2}})


def test_typeddict_total_true_e2e_optional_and_required_keys():
    o = from_env(EnvContTT, {"td": {"a": 1, "ro": 9}})
    assert o.td == {"a": 1, "ro": 9}

    d = o.to_dict()
    assert d == {"td": {"a": 1, "ro": 9}}

    with pytest.raises(Exception):
        from_env(EnvContTT, {"td": {"a": 1}})   # missing ro


def test_typeddict_all_required_e2e_inline_path():
    o = from_env(EnvContAllReq, {"td": {"x": 1, "y": "ok"}})
    assert o.td == {"x": 1, "y": "ok"}

    d = o.to_dict()
    assert d == {"td": {"x": 1, "y": "ok"}}

    with pytest.raises(Exception):
        from_env(EnvContAllReq, {"td": {"x": 1}})  # missing y


def test_union_codegen_cache_nested_union_roundtrip_and_dump_error():
    class MyClass(EnvWizard):
        class _(EnvWizard.Meta):
            unsafe_parse_dataclass_in_union = True

        complex_tp: 'list[int | Sub2] | list[int | str]'

    # First: pick the arm list[int|Sub2]
    o1 = from_env(MyClass, {"complex_tp": [{"my_float": "123."}, 7]})
    assert o1 == MyClass(complex_tp=[Sub2(my_float=123.0), 7])

    d1 = o1.to_dict()
    assert d1 == {"complex_tp": [{"my_float": 123.0}, 7]}

    # Second: pick the other arm list[int|str]
    # If inner-union codegen caching is wrong, this is where it tends to misbehave.
    o2 = from_env(MyClass, {"complex_tp": ["hello", 9]})
    assert o2 == MyClass(complex_tp=["hello", 9])

    d2 = o2.to_dict()
    assert d2 == {"complex_tp": ["hello", 9]}

    o1.complex_tp = '123'
    with pytest.raises(ParseError, match=r"Failed to dump field `complex_tp` in class `.*MyClass`"):
        _ = o1.to_dict()



class Condition:

    __dcw_condition__ = True
    __slots__ = (
        'op',
        'val',
        't_or_f',
        '_wrapped',
    )

    def __init__(self, operator, value):
        self.op = operator
        self.val = value
        self.t_or_f = operator in {'+', '!'}

    def __str__(self):
        return f"{self.op} {self.val!r}"

    def evaluate(self, other) -> bool:  # pragma: no cover
        # Optionally support runtime evaluation of the condition
        operators = {
            "==": lambda a, b: a == b,
            "!=": lambda a, b: a != b,
            "<": lambda a, b: a < b,
            "<=": lambda a, b: a <= b,
            ">": lambda a, b: a > b,
            ">=": lambda a, b: a >= b,
            "is": lambda a, b: a is b,
            "is not": lambda a, b: a is not b,
            "+": lambda a, _: True if a else False,
            "!": lambda a, _: not a,
        }
        return operators[self.op](other, self.val)


# Aliases for conditions

# noinspection PyPep8Naming
def EQ(value): return Condition("==", value)
# noinspection PyPep8Naming
def NE(value): return Condition("!=", value)
# noinspection PyPep8Naming
def LT(value): return Condition("<", value)
# noinspection PyPep8Naming
def LE(value): return Condition("<=", value)
# noinspection PyPep8Naming
def GT(value): return Condition(">", value)
# noinspection PyPep8Naming
def GE(value): return Condition(">=", value)
# noinspection PyPep8Naming
def IS(value): return Condition("is", value)
# noinspection PyPep8Naming
def IS_NOT(value): return Condition("is not", value)
# noinspection PyPep8Naming
def IS_TRUTHY(): return Condition("+", None)
# noinspection PyPep8Naming
def IS_FALSY(): return Condition("!", None)


# noinspection PyPep8Naming
def SkipIf(condition):
    """
    Mark a condition to be used as a skip directive during serialization.
    """
    condition._wrapped = True  # Set a marker attribute
    return condition


# Convenience alias, to skip serializing field if value is None
SkipIfNone = SkipIf(IS(None))



import re


def to_camel_case(string: str) -> str:
    """
    Convert a string to Camel Case.

    Examples::

        >>> to_camel_case("device_type")
        'deviceType'

    """
    string = replace_multi_with_single(
        string.replace('-', '_').replace(' ', '_'))

    return string[0].lower() + re.sub(
        r"_(.)", lambda m: m.group(1).upper(), string[1:])


def to_pascal_case(string):
    """
    Converts a string to Pascal Case (also known as "Upper Camel Case")

    Examples::

        >>> to_pascal_case("device_type")
        'DeviceType'

    """
    string = replace_multi_with_single(
        string.replace('-', '_').replace(' ', '_'))

    return string[0].upper() + re.sub(
        r"_(.)", lambda m: m.group(1).upper(), string[1:])


def to_lisp_case(string: str) -> str:
    """
    Make a hyphenated, lowercase form from the expression in the string.

    Example::

        >>> to_lisp_case("DeviceType")
        'device-type'

    """
    string = string.replace('_', '-').replace(' ', '-')
    # Short path: the field is already lower-cased, so we don't need to handle
    # for camel or title case.
    if string.islower():
        return replace_multi_with_single(string, '-')

    result = re.sub(
        r'((?!^)(?<!-)[A-Z][a-z]+|(?<=[a-z0-9])[A-Z])', r'-\1', string)

    return replace_multi_with_single(result.lower(), '-')


def to_snake_case(string: str) -> str:
    """
    Make an underscored, lowercase form from the expression in the string.

    Example::

        >>> to_snake_case("DeviceType")
        'device_type'

    """
    string = string.replace('-', '_').replace(' ', '_')
    # Short path: the field is already lower-cased, so we don't need to handle
    # for camel or title case.
    if string.islower():
        return replace_multi_with_single(string)

    result = re.sub(
        r'((?!^)(?<!_)[A-Z][a-z]+|(?<=[a-z0-9])[A-Z])', r'_\1', string)

    return replace_multi_with_single(result.lower())


def replace_multi_with_single(string: str, char='_') -> str:
    """
    Replace multiple consecutive occurrences of `char` with a single one.
    """
    rep = char + char
    while rep in string:
        string = string.replace(rep, char)

    return string


# Note: this is the initial helper function I came up with. This doesn't use
# regex for the string transformation, so it's actually faster than the
# implementation above. However, I do prefer the implementation with regex,
# because its a lot cleaner and more simple than this implementation.
# def to_snake_case_old(string: str):
#     """
#     Make an underscored, lowercase form from the expression in the string.
#     """
#     if len(string) < 2:
#         return string or ''
#
#     string = string.replace('-', '_')
#
#     if string.islower():
#         return replace_multi_with_single(string)
#
#     start_idx = 0
#
#     parts = []
#     for i, c in enumerate(string):
#         c: str
#         if c.isupper():
#             try:
#                 next_lower = string[i + 1].islower()
#             except IndexError:
#                 if string[i - 1].islower():
#                     parts.append(string[start_idx:i])
#                     parts.append(c)
#                 else:
#                     parts.append(string[start_idx:])
#                 break
#             else:
#                 if i == 0:
#                     continue
#
#                 if string[i - 1].islower():
#                     parts.append(string[start_idx:i])
#                     start_idx = i
#
#                 elif next_lower:
#                     parts.append(string[start_idx:i])
#                     start_idx = i
#     else:
#         parts.append(string[start_idx:i + 1])
#
#     result = '_'.join(parts).lower()
#
#     return replace_multi_with_single(result)



"""
Contains implementations for Abstract Base Classes
"""
import json
from abc import ABC, abstractmethod
from typing import AnyStr, ClassVar, TypeVar

from ._models import Extras, TypeInfo
from ._type_def import Encoder, JSONObject, ListOfJSONObject

# Create a generic variable that can be 'AbstractEnvWizard', or any subclass.
E = TypeVar('E', bound='AbstractEnvWizard')

# Create a generic variable that can be 'AbstractJSONWizard', or any subclass.
W = TypeVar('W', bound='AbstractJSONWizard')


class AbstractEnvWizard(ABC):
    """
    Abstract class that defines the methods a sub-class must implement at a
    minimum to be considered a "true" Environment Wizard.
    """
    __slots__ = ()

    # Extends the `__annotations__` attribute to return only the field
    # names of the `EnvWizard` subclass.
    #
    # .. NOTE::
    #    This excludes fields marked as ``ClassVar``, or ones which are
    #    not type-annotated.
    __field_names__: ClassVar[tuple[str, ...]]

    def raw_dict(self: E) -> JSONObject:
        """
        Same as ``__dict__``, but only returns values for fields defined
        on the `EnvWizard` instance. See :attr:`__field_names__` for more info.

        .. NOTE::
           The values in the returned dictionary object are not needed to be
           JSON serializable. Use :meth:`to_dict` if this is required.
        """

    @abstractmethod
    def to_dict(self: E) -> JSONObject:
        """
        Converts an instance of a `EnvWizard` subclass to a Python dictionary
        object that is JSON serializable.
        """

    @abstractmethod
    def to_json(self: E, indent=None) -> str:
        """
        Converts an instance of a `EnvWizard` subclass to a JSON `string`
        representation.
        """


class AbstractJSONWizard(ABC):
    """
    Abstract class that defines the methods a sub-class must implement at a
    minimum to be considered a "true" JSON Wizard.

    In particular, these are the abstract methods which - if correctly
    implemented - will allow a concrete sub-class (ideally a dataclass) to
    be properly loaded from, and serialized to, JSON.

    """

    @classmethod
    @abstractmethod
    def from_json(cls: type[W], string: AnyStr) -> W | list[W]:
        """
        Converts a JSON `string` to an instance of the dataclass, or a list of
        the dataclass instances.
        """

    @classmethod
    @abstractmethod
    def from_list(cls: type[W], o: ListOfJSONObject) -> list[W]:
        """
        Converts a Python `list` object to a list of the dataclass instances.
        """

    @classmethod
    @abstractmethod
    def from_dict(cls: type[W], o: JSONObject) -> W:
        """
        Converts a Python `dict` object to an instance of the dataclass.
        """

    @abstractmethod
    def to_dict(self: W) -> JSONObject:
        """
        Converts the dataclass instance to a Python dictionary object that is
        JSON serializable.
        """

    @abstractmethod
    def to_json(self: W, *,
                encoder: Encoder = json.dumps,
                indent=None,
                **encoder_kwargs) -> str:
        """
        Converts the dataclass instance to a JSON `string` representation.
        """

    @classmethod
    @abstractmethod
    def list_to_json(cls: type[W],
                     instances: list[W],
                     encoder: Encoder = json.dumps,
                     **encoder_kwargs) -> str:
        """
        Converts a ``list`` of dataclass instances to a JSON `string`
        representation.
        """


class AbstractLoaderGenerator(ABC):
    """
    Abstract code generator which defines helper methods to generate the
    code for deserializing an object `o` of a given annotated type into
    the corresponding dataclass field during dynamic function construction.
    """
    __slots__ = ()

    @staticmethod
    @abstractmethod
    def transform_json_field(string: str) -> str:
        """
        Transform a JSON field name (which will typically be camel-cased)
        into the conventional format for a dataclass field name
        (which will ideally be snake-cased).
        """

    @staticmethod
    @abstractmethod
    def is_none(tp: TypeInfo, extras: Extras) -> str:
        """
        Generate the condition to determine if a value is None.
        """

    @staticmethod
    @abstractmethod
    def load_fallback(tp: TypeInfo, extras: Extras) -> str:
        """
        Generate code for the fallback load handler
        when no specialized type matches.

        The default fallback implementation is typically
        an identity / passthrough, but subclasses may
        override this behavior.
        """

    @staticmethod
    @abstractmethod
    def load_to_str(tp: TypeInfo, extras: Extras) -> str:
        """
        Generate code to load a value into a string field.
        """

    @staticmethod
    @abstractmethod
    def load_to_int(tp: TypeInfo, extras: Extras) -> str:
        """
        Generate code to load a value into an integer field.
        """

    @staticmethod
    @abstractmethod
    def load_to_float(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into a float field.
        """

    @staticmethod
    @abstractmethod
    def load_to_bool(_: str, extras: Extras) -> str:
        """
        Generate code to load a value into a boolean field.
        Adds a helper function `as_bool` to the local context.
        """

    @staticmethod
    @abstractmethod
    def load_to_bytes(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into a bytes field.
        """

    @staticmethod
    @abstractmethod
    def load_to_bytearray(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into a bytearray field.
        """

    @staticmethod
    @abstractmethod
    def load_to_none(tp: TypeInfo, extras: Extras) -> str:
        """
        Generate code to load a value into a None.
        """

    @staticmethod
    @abstractmethod
    def load_to_literal(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to confirm a value is equivalent to one
        of the provided literals.
        """

    @classmethod
    @abstractmethod
    def load_to_union(cls, tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into a `Union[X, Y, ...]`
        (one of [X, Y, ...] possible types)
        """

    @staticmethod
    @abstractmethod
    def load_to_enum(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into an Enum field.
        """

    @staticmethod
    @abstractmethod
    def load_to_uuid(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into a UUID field.
        """

    @staticmethod
    @abstractmethod
    def load_to_iterable(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into an iterable field
        (list, set, etc.).
        """

    @staticmethod
    @abstractmethod
    def load_to_tuple(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into a tuple field.
        """

    @classmethod
    @abstractmethod
    def load_to_named_tuple(
            cls, tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into a named tuple field.
        """

    @classmethod
    @abstractmethod
    def load_to_named_tuple_untyped(
            cls, tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into an untyped named tuple.
        """

    @staticmethod
    @abstractmethod
    def load_to_dict(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into a dictionary field.
        """

    @staticmethod
    @abstractmethod
    def load_to_defaultdict(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into a defaultdict field.
        """

    @staticmethod
    @abstractmethod
    def load_to_typed_dict(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into a typed dictionary field.
        """

    @staticmethod
    @abstractmethod
    def load_to_decimal(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into a Decimal field.
        """

    @staticmethod
    @abstractmethod
    def load_to_path(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into a Path field.
        """

    @staticmethod
    @abstractmethod
    def load_to_date(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into a date field.
        """

    @staticmethod
    @abstractmethod
    def load_to_datetime(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into a datetime field.
        """

    @staticmethod
    @abstractmethod
    def load_to_time(tp: TypeInfo, extras: Extras) -> str:
        """
        Generate code to load a value into a time field.
        """

    @staticmethod
    @abstractmethod
    def load_to_timedelta(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into a timedelta field.
        """

    @staticmethod
    def load_to_dataclass(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to load a value into a `dataclass` type field.
        """

    @classmethod
    @abstractmethod
    def load_dispatcher_for_annotation(cls,
                                       tp: TypeInfo,
                                       extras: Extras) -> str | TypeInfo:
        """
        Resolve the load dispatcher for a given annotation type.

        Returns either a string reference to a dispatcher or a TypeInfo object,
        depending on how the annotation is handled.
        """


class AbstractDumperGenerator(ABC):
    """
    Abstract code generator which defines helper methods to generate the
    code for deserializing an object `o` of a given annotated type into
    the corresponding dataclass field during dynamic function construction.
    """
    __slots__ = ()

    @staticmethod
    @abstractmethod
    def transform_dataclass_field(string: str) -> str:
        """
        Transform a dataclass field name (which will ideally be snake-cased)
        into the conventional format for a JSON field name.
        """

    @staticmethod
    @abstractmethod
    def dump_fallback(tp: TypeInfo, extras: Extras) -> str:
        """
        Generate code for the fallback dump handler when no
        specialized type matches.

        The default fallback implementation is typically
        an identity / passthrough, but subclasses may
        override this behavior.
        """

    @staticmethod
    @abstractmethod
    def dump_from_str(tp: TypeInfo, extras: Extras) -> str:
        """
        Generate code to dump a value from a string field.
        """

    @staticmethod
    @abstractmethod
    def dump_from_int(tp: TypeInfo, extras: Extras) -> str:
        """
        Generate code to dump a value from an integer field.
        """

    @staticmethod
    @abstractmethod
    def dump_from_float(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from a float field.
        """

    @staticmethod
    @abstractmethod
    def dump_from_bool(_: str, extras: Extras) -> str:
        """
        Generate code to dump a value from a boolean field.
        """

    @staticmethod
    @abstractmethod
    def dump_from_bytes(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from a bytes field.
        """

    @staticmethod
    @abstractmethod
    def dump_from_bytearray(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from a bytearray field.
        """

    @staticmethod
    @abstractmethod
    def dump_from_none(tp: TypeInfo, extras: Extras) -> str:
        """
        Generate code to dump a value from a None.
        """

    @staticmethod
    @abstractmethod
    def dump_from_literal(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a literal.
        """

    @classmethod
    @abstractmethod
    def dump_from_union(cls, tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from a `Union[X, Y, ...]`
        (one of [X, Y, ...] possible types)
        """

    @staticmethod
    @abstractmethod
    def dump_from_enum(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from an Enum field.
        """

    @staticmethod
    @abstractmethod
    def dump_from_uuid(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from a UUID field.
        """

    @staticmethod
    @abstractmethod
    def dump_from_iterable(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from an iterable field
        (list, set, etc.).
        """

    @staticmethod
    @abstractmethod
    def dump_from_tuple(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from a tuple field.
        """

    @staticmethod
    @abstractmethod
    def dump_from_named_tuple(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from a named tuple field.
        """

    @classmethod
    @abstractmethod
    def dump_from_named_tuple_untyped(
            cls, tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from an untyped named tuple.
        """

    @staticmethod
    @abstractmethod
    def dump_from_dict(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from a dictionary field.
        """

    @staticmethod
    @abstractmethod
    def dump_from_defaultdict(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from a defaultdict field.
        """

    @staticmethod
    @abstractmethod
    def dump_from_typed_dict(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from a typed dictionary field.
        """

    @staticmethod
    @abstractmethod
    def dump_from_decimal(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from a Decimal field.
        """

    @staticmethod
    @abstractmethod
    def dump_from_path(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from a Decimal field.
        """

    @staticmethod
    @abstractmethod
    def dump_from_datetime(tp: TypeInfo, extras: Extras) -> str:
        """
        Generate code to dump a value from a datetime field.
        """

    @staticmethod
    @abstractmethod
    def dump_from_time(tp: TypeInfo, extras: Extras) -> str:
        """
        Generate code to dump a value from a time field.
        """

    @staticmethod
    @abstractmethod
    def dump_from_date(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from a date field.
        """

    @staticmethod
    @abstractmethod
    def dump_from_timedelta(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from a timedelta field.
        """

    @staticmethod
    def dump_from_dataclass(tp: TypeInfo, extras: Extras) -> str | TypeInfo:
        """
        Generate code to dump a value from a `dataclass` type field.
        """

    @classmethod
    @abstractmethod
    def dump_dispatcher_for_annotation(cls,
                                       tp: TypeInfo,
                                       extras: Extras) -> str | TypeInfo:
        """
        Resolve the dump dispatcher for a given annotation type.

        Returns either a string reference to a dispatcher or a TypeInfo
        object, depending on how the annotation is handled.
        """



__all__ = ['normalize',
           'possible_json_keys',
           'possible_env_vars',
           'repl_or_with_union']

from ..enums import EnvKeyStrategy

def normalize(string: str) -> str: ...
def possible_json_keys(field: str) -> list: ...
def possible_env_vars(field: str, lookup_strat: EnvKeyStrategy) -> list: ...
def to_camel_case(string: str) -> str: ...
def to_pascal_case(string): ...
def to_lisp_case(string: str) -> str: ...
def to_snake_case(string: str) -> str: ...
def repl_or_with_union(s: str): ...



import json
from typing import AnyStr, Collection, Callable, Protocol, dataclass_transform

from .abstractions import AbstractJSONWizard, W
from .bases_meta import BaseJSONWizardMeta, V1HookFn
from .enums import LetterCase
from .type_def import Decoder, Encoder, JSONObject, ListOfJSONObject


# A handy alias in case it comes in useful to anyone :)
JSONWizard = JSONSerializable


class SerializerHookMixin(Protocol):
    @classmethod
    def _pre_from_dict(cls: type[W], o: JSONObject) -> JSONObject:
        """
        Optional hook that runs before the dataclass instance is
        loaded, and before it is converted from a dictionary object
        via :meth:`from_dict`.

        To override this, subclasses need to implement this method.
        A simple example is shown below:

        >>> from dataclasses import dataclass
        >>> from dataclass_wizard import JSONWizard
        >>> from dataclass_wizard.type_def import JSONObject
        >>>
        >>>
        >>> @dataclass
        >>> class MyClass(JSONWizard):
        >>>     a_bool: bool
        >>>
        >>>     @classmethod
        >>>     def _pre_from_dict(cls, o: JSONObject) -> JSONObject:
        >>>         # o = o.copy()  # Copying the `dict` object is optional
        >>>         o['a_bool'] = True  # Add a new key/value pair
        >>>         return o
        >>>
        >>> c = MyClass.from_dict({})
        >>> assert c == MyClass(a_bool=True)
        """
        ...

    def _pre_dict(self):
        # noinspection PyDunderSlots, PyUnresolvedReferences
        """
        Optional hook that runs before the dataclass instance is processed and
        before it is converted to a dictionary object via :meth:`to_dict`.

        To override this, subclasses need to extend from :class:`DumpMixIn`
        and implement this method. A simple example is shown below:

        >>> from dataclasses import dataclass
        >>> from dataclass_wizard import JSONWizard
        >>>
        >>>
        >>> @dataclass
        >>> class MyClass(JSONWizard):
        >>>     my_str: str
        >>>
        >>>     def _pre_dict(self):
        >>>         self.my_str = self.my_str.swapcase()
        >>>
        >>> assert MyClass('test').to_dict() == {'myStr': 'TEST'}
        """
        ...


class JSONWizardImpl(AbstractJSONWizard, SerializerHookMixin):
    """
    Mixin class to allow a `dataclass` sub-class to be easily converted
    to and from JSON.

    """
    __slots__ = ()

    class Meta(BaseJSONWizardMeta):
        """
        Inner meta class that can be extended by sub-classes for additional
        customization with the JSON load / dump process.
        """
        __slots__ = ()

        # Class attribute to enable detection of the class type.
        __is_inner_meta__ = True

        def __init_subclass__(cls):
            # Set the `__init_subclass__` method here, so we can ensure it
            # doesn't run for the `JSONSerializable.Meta` class.
            ...

    @classmethod
    def register_type(cls, tp: type, *,
                      load: V1HookFn | None = None,
                      dump: V1HookFn | None = None,
                      mode: str | None = None) -> None:
        ...

    @classmethod
    def from_json(cls: type[W], string: AnyStr, *,
                  decoder: Decoder = json.loads,
                  **decoder_kwargs) -> W | list[W]:
        """
        Converts a JSON `string` to an instance of the dataclass, or a list of
        the dataclass instances.
        """
        ...

    @classmethod
    def from_list(cls: type[W], o: ListOfJSONObject) -> list[W]:
        """
        Converts a Python `list` object to a list of the dataclass instances.
        """
        # alias: fromlist(cls, o)
        ...

    @classmethod
    def from_dict(cls: type[W], o: JSONObject) -> W:
        # alias: fromdict(cls, o)
        ...

    def to_dict(self: W,
                *,
                dict_factory=dict,
                exclude: Collection[str] | None = None,
                skip_defaults: bool | None = None,
                ) -> JSONObject:
        """
        Converts the dataclass instance to a Python dictionary object that is
        JSON serializable.

        Example usage:

          @dataclass
          class C(JSONWizard):
              x: int
              y: int
              z: bool = True

          c = C(1, 2, True)
          assert c.to_dict(skip_defaults=True) == {'x': 1, 'y': 2}

        If given, 'dict_factory' will be used instead of built-in dict.
        The function applies recursively to field values that are
        dataclass instances. This will also look into built-in containers:
        tuples, lists, and dicts.
        """
        # alias: asdict(self)
        ...

    def to_json(self: W, *,
                encoder: Encoder = json.dumps,
                **encoder_kwargs) -> str:
        """
        Converts the dataclass instance to a JSON `string` representation.
        """
        ...

    @classmethod
    def list_to_json(cls: type[W],
                     instances: list[W],
                     encoder: Encoder = json.dumps,
                     **encoder_kwargs) -> str:
        """
        Converts a ``list`` of dataclass instances to a JSON `string`
        representation.
        """
        ...

    def __init_subclass__(cls,
                          str: bool = False,
                          debug: bool | str | int = False,
                          case: KeyCase | str | None = None,
                          dump_case: KeyCase | str | None = None,
                          load_case: KeyCase | str | None = None,
                          _key_transform: LetterCase | str | None = None,
                          _apply_dataclass: bool = True,
                          **dc_kwargs):
        """
        Checks for optional settings and flags that may be passed in by the
        sub-class, and calls the Meta initializer when :class:`Meta` is sub-classed.

        :param str: True to add a default ``__str__`` method to the subclass.
        :param debug: True to enable debug mode and setup logging, so that
          this library's DEBUG (and above) log messages are visible. If
          ``debug`` is a string or integer, it is assumed to be the desired
          "minimum logging level", and will be passed to ``logging.setLevel``.

        """
        ...


@dataclass_transform()
class DataclassWizard(JSONWizardImpl):
    ...


class JSONPyWizard(JSONWizardImpl):
    """Helper for JSONWizard that ensures dumping to JSON keeps keys as-is."""


class JSONSerializable(JSONWizardImpl): ...


def _str_fn() -> Callable[[W], str]:
    """
    Converts the dataclass instance to a *prettified* JSON string
    representation, when the `str()` method is invoked.
    """
    ...




__all__ = ['JSONListWizard',
           'JSONFileWizard',
           'TOMLWizard',
           'YAMLWizard']

import json
from os import PathLike
from typing import AnyStr, TextIO, BinaryIO, Union, TypeAlias

from .abstractions import W
from .enums import LetterCase
from .models import Container
from .serial_json import JSONSerializable, SerializerHookMixin
from .type_def import (T, ListOfJSONObject,
                       Encoder, Decoder, FileDecoder, FileEncoder, ParseFloat)


# A type that can be string or `path.Path`
# https://stackoverflow.com/a/78070015/10237506
# A type that can be string, bytes, or `PathLike`
FileType: TypeAlias = str | bytes | PathLike


class JSONListWizard(JSONSerializable, str=False):

    @classmethod
    def from_json(cls: type[W], string: AnyStr, *,
                  decoder: Decoder = json.loads,
                  **decoder_kwargs) -> W | Container[W]:

        ...

    @classmethod
    def from_list(cls: type[W], o: ListOfJSONObject) -> Container[W]:
        ...


class JSONFileWizard(SerializerHookMixin):

    @classmethod
    def from_json_file(cls: type[T], file: FileType, *,
                       decoder: FileDecoder = json.load,
                       **decoder_kwargs) -> T | list[T]:
        ...

    def to_json_file(self: T, file: FileType, mode: str = 'w',
                     encoder: FileEncoder = json.dump,
                     **encoder_kwargs) -> None:
        ...


class TOMLWizard(SerializerHookMixin):

    def __init_subclass__(cls, key_transform=LetterCase.NONE):
        ...

    @classmethod
    def from_toml(cls: type[T],
                  string_or_stream: AnyStr | BinaryIO, *,
                  decoder: Decoder | None = None,
                  header: str = 'items',
                  parse_float: ParseFloat = float) -> T | list[T]:
        ...

    @classmethod
    def from_toml_file(cls: type[T], file: FileType, *,
                       decoder: FileDecoder | None = None,
                       header: str = 'items',
                       parse_float: ParseFloat = float) -> T | list[T]:
        ...

    def to_toml(self: T,
                /,
                *encoder_args,
                encoder: Encoder | None = None,
                multiline_strings: bool = False,
                indent: int = 4) -> AnyStr:
        ...

    def to_toml_file(self: T, file: FileType, mode: str = 'wb',
                     encoder: FileEncoder | None = None,
                     multiline_strings: bool = False,
                     indent: int = 4) -> None:
        ...

    @classmethod
    def list_to_toml(cls: type[T],
                     instances: list[T],
                     header: str = 'items',
                     encoder: Encoder | None = None,
                     **encoder_kwargs) -> AnyStr:
        ...


class YAMLWizard(SerializerHookMixin):

    def __init_subclass__(cls, key_transform=LetterCase.LISP):
        ...

    @classmethod
    def from_yaml(cls: type[T],
                  string_or_stream: AnyStr | TextIO | BinaryIO, *,
                  decoder: Decoder | None = None,
                  **decoder_kwargs) -> T | list[T]:
        ...

    @classmethod
    def from_yaml_file(cls: type[T], file: FileType, *,
                       decoder: FileDecoder | None = None,
                       **decoder_kwargs) -> T | list[T]:
        ...

    def to_yaml(self: T, *,
                encoder: Encoder | None = None,
                **encoder_kwargs) -> AnyStr:
        ...

    def to_yaml_file(self: T, file: FileType, mode: str = 'w',
                     encoder: FileEncoder | None = None,
                     **encoder_kwargs) -> None:
        ...

    @classmethod
    def list_to_yaml(cls: type[T],
                     instances: list[T],
                     encoder: Encoder | None = None,
                     **encoder_kwargs) -> AnyStr:
        ...



import json
from collections.abc import Collection, Mapping
from dataclasses import Field, InitVar, dataclass
from typing import (
    Callable,
    ClassVar,
    NotRequired,
    TypedDict,
    TypeVar,
    dataclass_transform,
)

from ._bases import AbstractEnvMeta
from ._bases_meta import BaseEnvWizardMeta, HookFn
from ._loaders import LoadMixin as V1LoadMixIn
from ._models import Extras, TypeInfo
from ._type_def import ENV_META, Encoder, JSONObject, T, Unpack

E_ = TypeVar('E_', bound=EnvWizard)
E = type[E_]


class EnvInit(TypedDict, total=False):
    mapping: NotRequired[Mapping[str, str]]
    file: NotRequired[str | list[str] | bool]
    prefix: NotRequired[str]
    secrets_dir: NotRequired[str | list[str]]


def env_config(**kw: Unpack[EnvInit]) -> EnvInit:
    return kw


@dataclass_transform(kw_only_default=True)
@dataclass()
class EnvWizard:
    __slots__ = ()

    # noinspection PyDataclass
    __env__: InitVar[EnvInit | None] = None

    class Meta(BaseEnvWizardMeta):
        """
        Inner meta class that can be extended by sub-classes for additional
        customization with the environment load process.
        """
        __slots__ = ()
        __is_inner_meta__ = True
        def __init_subclass__(cls): ...

    __field_names__: ClassVar[tuple[str, ...]]

    def __init_subclass__(cls,
                          debug: bool = False,
                          _apply_dataclass=True,
                          **dc_kwargs): ...

    @classmethod
    def register_type(cls, tp: type, *,
                      load: HookFn | None = None,
                      dump: HookFn | None = None,
                      mode: str | None = None) -> None:
        ...

    def raw_dict(self: E_) -> JSONObject: ...

    def to_dict(self: E_,
                *,
                dict_factory=dict,
                exclude: Collection[str] | None = None,
                skip_defaults: bool | None = None,
                ) -> JSONObject:
        """
        Converts the dataclass instance to a Python dictionary object that is
        JSON serializable.

        Example usage:

          class C(EnvWizard):
              x: int
              y: int
              z: bool = True

          c = C(1, 2, True)
          assert c.to_dict(skip_defaults=True) == {'x': 1, 'y': 2}

        If given, 'dict_factory' will be used instead of built-in dict.
        The function applies recursively to field values that are
        dataclass instances. This will also look into built-in containers:
        tuples, lists, and dicts.
        """
        # alias: asdict(self)
        ...

    def to_json(self: E_, *,
                encoder: Encoder = json.dumps,
                **encoder_kwargs) -> str:
        """
        Converts the dataclass instance to a JSON `string` representation.
        """
        ...

def load_func_for_dataclass(
        cls: E,
        extras: Extras | None = None,
        loader_cls=None,
        base_meta_cls: ENV_META = AbstractEnvMeta,
) -> Callable[[JSONObject], T] | None: ...

def _add_missing_var(missing_vars: dict | None, name, env_prefix, var_name, tp): ...

def generate_field_code(cls_loader: LoadMixin,
                        extras: Extras,
                        field: Field,
                        field_i: int) -> str | TypeInfo: ...

def re_raise(e, cls, o, fields, field, value): ...

class LoadMixin(V1LoadMixIn): ...



import json
from collections.abc import Collection
from typing import (
    Any,
    AnyStr,
    Callable,
    Protocol,
    Self,
    dataclass_transform,
)

from ._abstractions import AbstractJSONWizard, W
from ._bases_meta import BaseJSONWizardMeta
from ._type_def import Decoder, Encoder, JSONObject, ListOfJSONObject
from .enums import KeyCase

def first_declared_attr_in_mro(cls: type, name: str) -> Callable | Any | None: ...
def set_from_dict_and_to_dict_if_needed(cls: type) -> None: ...
def configure_wizard_class(cls: type,
                           str: bool = False,
                           debug: bool | int = False,
                           case: KeyCase | str | None = None,
                           dump_case: KeyCase | str | None = None,
                           load_case: KeyCase | str | None  = None):
    ...

class SerializerHookMixin(Protocol):
    @classmethod
    def _pre_from_dict(cls: type[Self], o: JSONObject) -> JSONObject:
        """
        Optional hook that runs before the dataclass instance is
        loaded, and before it is converted from a dictionary object
        via :meth:`from_dict`.

        To override this, subclasses need to implement this method.
        A simple example is shown below:

        >>> from dataclasses import dataclass
        >>> from dataclass_wizard import JSONWizard
        >>> from dataclass_wizard._type_def import JSONObject
        >>>
        >>>
        >>> @dataclass
        >>> class MyClass(JSONWizard):
        >>>     a_bool: bool
        >>>
        >>>     @classmethod
        >>>     def _pre_from_dict(cls, o: JSONObject) -> JSONObject:
        >>>         # o = o.copy()  # Copying the `dict` object is optional
        >>>         o['a_bool'] = True  # Add a new key/value pair
        >>>         return o
        >>>
        >>> c = MyClass.from_dict({})
        >>> assert c == MyClass(a_bool=True)
        """
        ...

    def _pre_to_dict(self: Self) -> Self:
        # noinspection PyDunderSlots, PyUnresolvedReferences
        """
        Optional hook that runs before the dataclass instance is processed and
        before it is converted to a dictionary object via :meth:`to_dict`.

        To override this, subclasses need to extend from :class:`DumpMixIn`
        and implement this method. A simple example is shown below:

        >>> from dataclasses import dataclass
        >>> from dataclass_wizard import JSONWizard
        >>>
        >>>
        >>> @dataclass
        >>> class MyClass(JSONWizard):
        >>>     my_str: str
        >>>
        >>>     def _pre_to_dict(self):
        >>>         self.my_str = self.my_str.swapcase()
        >>>         return self
        >>>
        >>> assert MyClass('test').to_dict() == {'myStr': 'TEST'}
        """
        ...


class _JSONWizardMixin(AbstractJSONWizard, SerializerHookMixin):
    """
    Mixin class to allow a `dataclass` sub-class to be easily converted
    to and from JSON.

    """

    class Meta(BaseJSONWizardMeta):
        """
        Inner meta class that can be extended by sub-classes for additional
        customization with the JSON load / dump process.
        """
        # Class attribute to enable detection of the class type.
        __is_inner_meta__ = True

    @classmethod
    def from_json(cls: type[W], string: AnyStr, *,
                  decoder: Decoder = json.loads,
                  **decoder_kwargs) -> W | list[W]:
        """
        Converts a JSON `string` to an instance of the dataclass, or a list of
        the dataclass instances.
        """
        ...

    @classmethod
    def from_list(cls: type[W], o: ListOfJSONObject) -> list[W]:
        """
        Converts a Python `list` object to a list of the dataclass instances.
        """
        # alias: fromlist(cls, o)
        ...

    @classmethod
    def from_dict(cls: type[W], o: JSONObject) -> W:
        # alias: fromdict(cls, o)
        ...

    def to_dict(self: W,
                *,
                dict_factory=dict,
                exclude: Collection[str] | None = None,
                skip_defaults: bool | None = None,
                ) -> JSONObject:
        """
        Converts the dataclass instance to a Python dictionary object that is
        JSON serializable.

        Example usage:

          @dataclass
          class C(JSONWizard):
              x: int
              y: int
              z: bool = True

          c = C(1, 2, True)
          assert c.to_dict(skip_defaults=True) == {'x': 1, 'y': 2}

        If given, 'dict_factory' will be used instead of built-in dict.
        The function applies recursively to field values that are
        dataclass instances. This will also look into built-in containers:
        tuples, lists, and dicts.
        """
        # alias: asdict(self)
        ...

    def to_json(self: W, *,
                encoder: Encoder = json.dumps,
                **encoder_kwargs) -> str:
        """
        Converts the dataclass instance to a JSON `string` representation.
        """
        ...

    @classmethod
    def list_to_json(cls: type[W],
                     instances: list[W],
                     encoder: Encoder = json.dumps,
                     **encoder_kwargs) -> str:
        """
        Converts a ``list`` of dataclass instances to a JSON `string`
        representation.
        """
        ...

@dataclass_transform()
class DataclassWizard(_JSONWizardMixin):
    class Meta(BaseJSONWizardMeta):
        ...
    @classmethod
    def from_dict(cls: type[W], o: JSONObject) -> W: ...
    @classmethod
    def from_list(cls: type[W], o: ListOfJSONObject) -> list[W]: ...
    @classmethod
    def from_json(cls: type[W], string: AnyStr, *, decoder: Decoder = ..., **decoder_kwargs) -> W | list[W]: ...
    def to_dict(self: W, *, dict_factory=..., exclude: Collection[str] | None = ..., skip_defaults: bool | None = ...) -> JSONObject: ...
    def to_json(self: W, *, encoder: Encoder = ..., **encoder_kwargs) -> str: ...
    @classmethod
    def list_to_json(cls: type[W], instances: list[W], encoder: Encoder = ..., **encoder_kwargs) -> str: ...

class JSONWizard(_JSONWizardMixin): ...

def _str_fn() -> Callable[[W], str]:
    """
    Converts the dataclass instance to a *prettified* JSON string
    representation, when the `str()` method is invoked.
    """
    ...



from collections.abc import Collection
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    NotRequired,
    Self,
    TypeAlias,
    TypedDict,
)

from ._type_def import META, DefFactory, T
from .conditions import Condition
from .patterns import PatternBase
from .utils._function_builder import FunctionBuilder

# Type for a string or a collection of strings.
_STR_COLLECTION: TypeAlias = str | Collection[str]
LEAF_TYPES: frozenset[type]
LEAF_TYPES_NO_BYTES: frozenset[type]
SEQUENCE_ORIGINS: frozenset[type]
MAPPING_ORIGINS: frozenset[type]

@dataclass(order=True)
class TypeInfo:
    # type origin (ex. `List[str]` -> `List`)
    origin: type
    # type arguments (ex. `Dict[str, int]` -> `(str, int)`)
    args: tuple[type, ...] | None = None
    # name of type origin (ex. `List[str]` -> 'list')
    name: str | None = None
    # index of iteration, *only* unique within the scope of a field assignment!
    i: int = 1
    # index of field within the dataclass, *guaranteed* to be unique.
    field_i: int = 1
    # prefix of value in assignment (prepended to `i`),
    # defaults to 'v' if not specified.
    prefix: str = 'v'
    # index / indices of assignment (ex. `2, 0 -> v1[2][0]`, *or* a string `"key" -> v4["key"]`)
    index: int | str | tuple[int | str, ...] | None = None
    # explicit value name (overrides prefix + index)
    val_name: str | None = None
    # indicates if we are currently in Optional,
    # e.g. `typing.Optional[...]` *or* `typing.Union[T, ...*T2, None]`
    in_optional: bool = False

    def replace(self, **changes) -> TypeInfo: ...
    @staticmethod
    def ensure_in_locals(extras: Extras, *tps: Callable | type, **name_to_tp: Callable[..., Any] | object) -> list[str]: ...
    def type_name(self, extras: Extras,
                  *, bound: type | None = None) -> str: ...
    def v(self) -> str: ...
    def v_for_def(self) -> str: ...
    def v_and_next(self) -> tuple[str, str, int]: ...
    def v_and_next_k_v(self) -> tuple[str, str, str, int]: ...
    def multi_wrap(self, extras, prefix='', *result, force=False) -> list[str]: ...
    def wrap(self, result: str,
             extras: Extras,
             force=False,
             prefix='',
             *, bound: type | None = None) -> Self: ...
    def wrap_builtin(self, bound: type, result: str, extras: Extras) -> Self: ...
    def wrap_dd(self, default_factory: DefFactory[T], result: str, extras: Extras) -> Self: ...
    def _wrap_inner(self, extras: Extras,
                    tp: type | DefFactory | None = None,
                    prefix: str = '',
                    is_builtin: bool = False,
                    force=False,
                    bound: type | None = None) -> str | None: ...


class Extras(TypedDict):
    """
    "Extra" config that can be used in the load / dump process.
    """
    config: META
    cls: type
    cls_name: str
    fn_gen: FunctionBuilder
    locals: dict[str, Any]
    pattern: NotRequired[PatternBase]
    recursion_guard: dict[Any, str]


def ensure_type_ref(extras: Extras, tp: type, *,
                    name: str | None = None,
                    prefix: str = '',
                    is_builtin: bool = False) -> str: ...

def finalize_skip_if(skip_if: Condition,
                     operand_1: str,
                     conditional: str) -> str:
    ...


def get_skip_if_condition(skip_if: Condition,
                          _locals: dict[str, Any],
                          operand_2: str | None = None,
                          condition_i: int | None = None,
                          condition_var: str = '_skip_if_') -> str | bool:
    ...



from __future__ import annotations

import json
import logging
from dataclasses import MISSING, dataclass

from ._bases_meta import BaseJSONWizardMeta, LoadMeta
from ._class_helper import call_meta_initializer_if_needed
from ._dumpers import asdict
from ._loaders import fromdict, fromlist
from ._log import enable_library_debug_logging
from ._type_def import UNSET, dataclass_transform
from .constants import PACKAGE_NAME
from .enums import KeyCase

# noinspection PyProtectedMember
from .utils._dataclass_compat import (
    dataclass_needs_refresh,
    set_new_attribute,
    str_pprint_fn,
)


def first_declared_attr_in_mro(cls, name):
    """First `name` found in MRO (excluding cls); else None."""
    for base in cls.__mro__[1:]:
        attr = base.__dict__.get(name, MISSING)
        if attr is not MISSING:
            return attr
    return None


def set_from_dict_and_to_dict_if_needed(cls):
    """
    Pin default dispatchers on subclasses.

    Codegen is lazy; if a base later gets a specialised
    `from_dict` / `to_dict`, subclasses would inherit it.
    Defining defaults in `cls.__dict__` blocks that.
    """
    cls.__dataclass_wizard_from_dict__ = UNSET
    cls.__dataclass_wizard_to_dict__ = UNSET

    if 'from_dict' not in cls.__dict__:
        inherited = first_declared_attr_in_mro(cls, 'from_dict')
        if getattr(inherited, '__func__', None) is fromdict:
            cls.from_dict = classmethod(fromdict)

    if 'to_dict' not in cls.__dict__:
        inherited = first_declared_attr_in_mro(cls, 'to_dict')
        if inherited is asdict:
            cls.to_dict = asdict


# noinspection PyShadowingBuiltins
def configure_wizard_class(cls,
                           str: bool = False,
                           debug: bool | str | int = False,
                           case: KeyCase | str | None = None,
                           dump_case: KeyCase | str | None = None,
                           load_case: KeyCase | str | None = None):
    load_meta_kwargs = {}

    if case is not None:
        load_meta_kwargs['case'] = case

    if dump_case is not None:
        load_meta_kwargs['dump_case'] = dump_case

    if load_case is not None:
        load_meta_kwargs['load_case'] = load_case

    if debug:
        # minimum logging level for logs by this library
        lvl = logging.DEBUG if isinstance(debug, bool) else debug
        # enable library logging
        enable_library_debug_logging(lvl)
        # set `debug` flag for the class's Meta
        load_meta_kwargs['debug'] = lvl

    if load_meta_kwargs:
        LoadMeta(**load_meta_kwargs).bind_to(cls)

    # Calls the Meta initializer when inner :class:`Meta` is sub-classed.
    call_meta_initializer_if_needed(cls)

    # Add a `__str__` method to the subclass, if needed
    if str:
        set_new_attribute(cls, '__str__', str_pprint_fn())

    # Add `from_dict` and `to_dict` methods to the subclass, if needed
    set_from_dict_and_to_dict_if_needed(cls)


@dataclass_transform()
class DataclassWizard:

    __slots__ = ()

    __dataclass_wizard_from_dict__ = UNSET
    __dataclass_wizard_to_dict__ = UNSET

    class Meta(BaseJSONWizardMeta):

        __slots__ = ()

        __is_inner_meta__ = True

        def __init_subclass__(cls):
            return cls._init_subclass()

    @classmethod
    def from_json(cls, string, *,
                  decoder=json.loads,
                  **decoder_kwargs):

        o = decoder(string, **decoder_kwargs)

        return fromdict(cls, o) if isinstance(o, dict) else fromlist(cls, o)

    from_list = classmethod(fromlist)

    from_dict = classmethod(fromdict)

    to_dict = asdict

    def to_json(self, *,
                encoder=json.dumps,
                **encoder_kwargs):

        return encoder(asdict(self), **encoder_kwargs)

    @classmethod
    def list_to_json(cls,
                     instances,
                     encoder=json.dumps,
                     **encoder_kwargs):

        list_of_dict = [asdict(o, cls=cls) for o in instances]

        return encoder(list_of_dict, **encoder_kwargs)

    # noinspection PyShadowingBuiltins
    def __init_subclass__(cls,
                          str=False,
                          debug=False,
                          case=None,
                          dump_case=None,
                          load_case=None,
                          _apply_dataclass=True,
                          **dc_kwargs):

        super().__init_subclass__()

        # skip classes provided by this library.
        if cls.__module__.startswith(f'{PACKAGE_NAME}.'):
            return

        # Apply the @dataclass decorator.
        if _apply_dataclass and dataclass_needs_refresh(cls):
            # noinspection PyArgumentList
            dataclass(cls, **dc_kwargs)

        configure_wizard_class(cls, str, debug, case, dump_case, load_case)


# noinspection PyAbstractClass
class JSONWizard(DataclassWizard):

    __slots__ = ()

    # noinspection PyShadowingBuiltins
    def __init_subclass__(cls,
                          str=False,
                          debug=False,
                          case=None,
                          dump_case=None,
                          load_case=None,
                          _apply_dataclass=False,
                          **_):

        super().__init_subclass__(str, debug, case, dump_case, load_case,
                                  _apply_dataclass)
